import os
import sys
from typing import List, Optional, Tuple

if os.name == "nt":
    os.environ["PYTHONUTF8"] = "1"

import logging
import signal
import time
import traceback
from collections import OrderedDict
from functools import wraps
from multiprocessing import Process
from pathlib import Path

import click
from rich import print
from rich.console import Console
from rich.prompt import Confirm, Prompt

from baidu import __version__
from baidu.app.account import Account, AccountManager
from baidu.baidupcs import BaiduPCSApi, BaiduPCSError
from baidu.baidupcs.inner import PcsRapidUploadInfo
from baidu.commands import cloud as _cloud
from baidu.commands import file_operators
from baidu.commands import share as _share
from baidu.commands.cat import cat as _cat
from baidu.commands.disk_usage import disk_usage
from baidu.commands.display import (
    display_user_info,
    display_user_infos,
)
from baidu.commands.download import (
    DEFAULT_CHUNK_SIZE,
    DEFAULT_CONCURRENCY,
    DEFAULT_DOWNLOADER,
    Downloader,
    DownloadParams,
)
from baidu.commands.download import (
    download as _download,
)
from baidu.commands.env import ACCOUNT_DATA_PATH, RAPIDUPLOADINFO_PATH
from baidu.commands.list_files import list_files
from baidu.commands.log import get_logger
from baidu.commands.play import DEFAULT_PLAYER, Player
from baidu.commands.play import play as _play
from baidu.commands.rapid_upload import (
    rapid_upload,
    rapid_upload_delete,
    rapid_upload_links,
    rapid_upload_list,
    rapid_upload_search,
)
from baidu.commands.search import search as _search
from baidu.commands.server import start_server
from baidu.commands.sifter import (
    ExcludeSifter,
    IncludeSifter,
    IsDirSifter,
    IsFileSifter,
)
from baidu.commands.sync import sync as _sync
from baidu.commands.upload import CPU_NUM, UploadType, from_tos
from baidu.commands.upload import upload as _upload
from baidu.common.event import keyboard_listener_start
from baidu.common.io import EncryptType
from baidu.common.net import random_avail_port
from baidu.common.path import join_path
from baidu.common.progress_bar import _progress, init_progress_bar

logger = get_logger(__name__)

DEBUG = logger.level == logging.DEBUG


def _teardown():
    logger.debug("`app`: _teardown: start")

    from baidu.common.event import _KEYBOARD_LISTENER

    if _KEYBOARD_LISTENER is not None:
        logger.debug("`app`: _teardown: reset old terminal")
        _KEYBOARD_LISTENER.set_normal_term()

    logger.debug("`app`: _teardown: end")

    # No use sys.exit() which only exits the main thread
    os._exit(1)


def _exit_progress_bar():
    if _progress.live._started:
        logger.debug("`app`: _exit_progress_bar: stop progress bar")

        print()
        # Stop _progress, otherwise terminal stdout will be contaminated
        _progress.stop()


def handle_signal(sign_num, frame):
    logger.debug("`app`: handle_signal: %s", sign_num)

    _exit_progress_bar()
    _teardown()


signal.signal(signal.SIGINT, handle_signal)


def handle_error(func):
    """Handle command error wrapper"""

    @wraps(func)
    def wrap(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except BaiduPCSError as err:
            _exit_progress_bar()

            logger.debug("`app`: BaiduPCSError: %s", traceback.format_exc())

            print(f"(v{__version__}) [bold red]ERROR[/bold red]: BaiduPCSError: {err}")
            if DEBUG:
                console = Console()
                console.print_exception()

        except Exception as err:
            _exit_progress_bar()

            logger.debug("`app`: System Error: %s", traceback.format_exc())

            print(f"(v{__version__}) [bold red]System ERROR[/bold red]: {err}")
            if DEBUG:
                console = Console()
                console.print_exception()

        finally:
            _exit_progress_bar()
            _teardown()

    return wrap


def _user_ids(ctx) -> Optional[List[int]]:
    """Select use_ids by their name probes"""

    am = ctx.obj.account_manager
    account_name_probes = ctx.obj.accounts

    user_ids = []
    for user_id, account in am._accounts.items():
        account_name = account.account_name
        for probe in account_name_probes:
            if probe in account_name:
                user_ids.append(user_id)
                break
    return user_ids


def _change_account(ctx, user_id: int):
    """Change recent account with `user_id`"""

    am = ctx.obj.account_manager
    am.su(user_id)


def multi_user_do(func):
    """Run command on multi accounts"""

    @wraps(func)
    def wrap(*args, **kwargs):
        ctx = args[0]
        user_ids = _user_ids(ctx)
        if not user_ids:
            return func(*args, **kwargs)

        am = ctx.obj.account_manager
        for user_id in user_ids:
            accout = am.who(user_id)
            if not accout:
                continue

            account_name = accout.account_name
            print(
                "[i yellow]@Do[/i yellow]: "
                f"account_name: [b]{account_name}[/b], "
                f"user_id: [b]{user_id}[/b]"
            )
            _change_account(ctx, user_id)
            func(*args, **kwargs)
            print()

    return wrap


def _recent_account(ctx) -> Optional[Account]:
    """Return recent user's `BaiduPCSApi`"""

    am = ctx.obj.account_manager
    account = am.who()
    if account:
        return account
    else:
        print("[italic red]No recent user, please adding or selecting one[/]")
        return None


def _recent_user_id_and_name(ctx) -> Tuple[Optional[int], Optional[str]]:
    account = _recent_account(ctx)
    user_id, user_name = None, None
    if account:
        user_id = account.user.user_id
        user_name = account.user.user_name
    return user_id, user_name


def _recent_api(ctx) -> Optional[BaiduPCSApi]:
    """Return recent user's `BaiduPCSApi`"""

    account = _recent_account(ctx)
    if account:
        return account.pcsapi()
    else:
        return None


def _pwd(ctx) -> str:
    """Return recent user's pwd"""

    am = ctx.obj.account_manager
    return am.pwd


def _encrypt_password(ctx) -> bytes:
    """Return recent user's encryption key"""

    account = _recent_account(ctx)
    if account:
        ep = account.encrypt_password
        return bytes(ep or "", encoding="utf-8")
    else:
        return b""


def _salt(ctx) -> bytes:
    """Return recent user's encryption key"""

    account = _recent_account(ctx)
    if account:
        salt = account.salt
        return bytes(salt or "", encoding="utf-8")
    else:
        return b""


def _rapiduploadinfo_file(ctx) -> str:
    """Return global `RapidUploadInfo`"""

    return ctx.obj.rapiduploadinfo_file


ALIAS = OrderedDict(
    **{
        # Account
        "w": "who",
        "uu": "updateuser",
        "su": "su",
        "ul": "userlist",
        "ua": "useradd",
        "ud": "userdel",
        "ep": "encryptpwd",
        "an": "accountname",
        # File Operations
        "l": "ls",
        "f": "search",
        "md": "mkdir",
        "mv": "move",
        "rn": "rename",
        "cp": "copy",
        "rm": "remove",
        "d": "download",
        "p": "play",
        "u": "upload",
        "sn": "sync",
        # Rapid Upload
        "rl": "rplist",
        "rs": "rpsearch",
        "rd": "rpdelete",
        "rp": "rp",
        # Share
        "S": "share",
        "sl": "shared",
        "cs": "cancelshared",
        "lsp": "listsharedpaths",
        "s": "save",
        # Cloud Task
        "a": "add",
        "t": "tasks",
        "ct": "cleartasks",
        "cct": "canceltasks",
        "sv": "server",
    }
)


class AliasedGroup(click.Group):
    def get_command(self, ctx, cmd_name):
        # As normal command name
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv

        # Check alias command name
        if cmd_name not in ALIAS:
            ctx.fail(f"No command: {cmd_name}")

        normal_cmd_name = ALIAS[cmd_name]
        return click.Group.get_command(self, ctx, normal_cmd_name)

    def list_commands(self, ctx):
        return self.commands


_APP_DOC = f"""BaiduPCS App v{__version__}

    \b
    如果第一次使用，你需要运行 `BaiduPCS-Py useradd` 添加 `bduss` 和 `cookies`。
    如何获取 `bduss` 和 `cookies` 见 https://github.com/PeterDing/BaiduPCS-Py#%E6%B7%BB%E5%8A%A0%E7%94%A8%E6%88%B7
    用 `BaiduPCS-Py {{command}} --help` 查看具体的用法。"""

_ALIAS_DOC = "Command 别名:\n\n\b\n" + "\n".join(
    [f"{alias: >3} : {cmd}" for alias, cmd in ALIAS.items()]
)


@click.group(cls=AliasedGroup, help=_APP_DOC, epilog=_ALIAS_DOC)
@click.option(
    "--account-data",
    "-ad",
    type=str,
    default=ACCOUNT_DATA_PATH,
    help="Account data file",
)
@click.option(
    "--rapiduploadinfo-file",
    "--rf",
    type=str,
    default=RAPIDUPLOADINFO_PATH,
    help="秒传 sqlite3 文件",
)
@click.option("--accounts", "-u", type=str, default=None, help="帐号名片段，用“,”分割")
@click.pass_context
def app(ctx, account_data, rapiduploadinfo_file, accounts):
    ctx.obj.account_manager = AccountManager.load_data(account_data)
    ctx.obj.rapiduploadinfo_file = str(Path(rapiduploadinfo_file).expanduser())
    ctx.obj.accounts = [] if accounts is None else accounts.split(",")


# Account
# {{{
@app.command()
@click.argument("user_id", type=int, default=None, required=False)
@click.option("--show-encrypt-password", "-K", is_flag=True, help="显示加密密码")
@click.pass_context
@handle_error
def who(ctx, user_id, show_encrypt_password):
    """显示当前用户的信息

    也可指定 `user_id`
    """

    am = ctx.obj.account_manager
    account = am.who(user_id)
    if account:
        display_user_info(account.user)
        if show_encrypt_password:
            encrypt_password = _encrypt_password(ctx)

            print(f"[red]encrypt password[/red]: {encrypt_password}")
    else:
        print("[italic red]No recent user, please adding or selecting one[/]")


@app.command()
@click.argument("user_ids", type=int, nargs=-1, default=None, required=False)
@click.option("--all", is_flag=True, help="更新所有用户信息")
@click.pass_context
@handle_error
def updateuser(ctx, user_ids, all):
    """更新用户信息 （默认更新当前用户信息）

    也可指定多个 `user_id`
    """

    am: AccountManager = ctx.obj.account_manager
    if not user_ids:
        if all:
            user_ids = [a.user.user_id for a in am.accounts]
        else:
            user_ids = [am._who]

    for user_id in user_ids:
        am.update(user_id)
        account = am.who(user_id)
        if account:
            display_user_info(account.user)

    am.save()


@app.command()
@click.argument("user_index", type=int, default=None, required=False)
@click.pass_context
@handle_error
def su(ctx, user_index):
    """切换当前用户"""

    am = ctx.obj.account_manager
    ls = sorted([(a.user, a.pwd, a.account_name) for a in am.accounts])
    display_user_infos(*ls, recent_user_id=am._who)

    if user_index:
        assert 0 < user_index < len(ls) + 1, f"Out range index {user_index}"

        i = user_index
        print(f"Select the user index: {i}")
    else:
        indexes = list(str(idx) for idx in range(1, len(ls) + 1))
        i = Prompt.ask("Select an user index", choices=indexes, default="")

        if not i:
            return

    user_id = ls[int(i) - 1][0].user_id
    am.su(user_id)
    am.save()


@app.command()
@click.pass_context
@handle_error
def userlist(ctx):
    """显示所有用户"""

    am = ctx.obj.account_manager
    ls = sorted([(a.user, a.pwd, a.account_name) for a in am.accounts])
    display_user_infos(*ls, recent_user_id=am._who)


@app.command()
@click.option(
    "--account_name", prompt="Account Name", hide_input=False, default="", help="账号名"
)
@click.option("--bduss", prompt="BDUSS", hide_input=True, default="", help="用户 BDUSS")
@click.option(
    "--cookies", prompt="Cookies", hide_input=True, default="", help="用户 Cookies"
)
@click.pass_context
@handle_error
def useradd(ctx, account_name, bduss, cookies):
    """添加一个用户并设置为当前用户"""

    if cookies:
        cookies = dict([c.split("=", 1) for c in cookies.split("; ")])
        bduss = bduss or cookies.get("BDUSS")
    else:
        cookies = {}
    if not bduss:
        raise ValueError("bduss must be specified or be included in cookie")
    account = Account.from_bduss(bduss, cookies=cookies, account_name=account_name)
    am = ctx.obj.account_manager
    am.add_account(account)
    am.su(account.user.user_id)
    am.save()


@app.command()
@click.pass_context
@handle_error
def userdel(ctx):
    """删除一个用户"""

    am = ctx.obj.account_manager
    ls = sorted([(a.user, a.pwd, a.account_name) for a in am.accounts])
    display_user_infos(*ls, recent_user_id=am._who)

    indexes = list(str(idx) for idx in range(1, len(ls) + 1))
    i = Prompt.ask("Delete an user index", choices=indexes, default="")
    if not i:
        return

    user_id = ls[int(i) - 1][0].user_id
    am.delete_account(user_id)
    am.save()

    print(f"Delete user {user_id}")


@app.command()
@click.option(
    "--encrypt-password",
    "--ep",
    prompt="encrypt-password",
    hide_input=True,
    help="加密密码，任意字符",
)
@click.option(
    "--salt",
    "-s",
    type=str,
    default=None,
    help="加密salt，不限字符 (^v0.5.17 后不使用)",
)
@click.pass_context
@handle_error
def encryptpwd(ctx, encrypt_password, salt):
    """设置加密密码"""

    assert len(encrypt_password) > 0, "No encrypt-password"

    am = ctx.obj.account_manager
    am.set_encrypt_password(encrypt_password, salt)
    am.save()


@app.command()
@click.option(
    "--account-name",
    "--an",
    prompt="Account Name",
    hide_input=False,
    help="设置账号名，用于指定运行帐号和显示",
)
@click.option("--user-id", "-i", type=int, help="指定用户id")
@click.pass_context
@handle_error
def accountname(ctx, account_name, user_id):
    """设置账号名"""

    am = ctx.obj.account_manager
    am.set_account_name(account_name, user_id)
    am.save()


@app.command()
@click.argument("remotedir", type=str, default="/", required=False)
@click.pass_context
@handle_error
def cd(ctx, remotedir):
    """切换当前工作目录"""

    am = ctx.obj.account_manager
    am.cd(remotedir)
    am.save()


@app.command()
@click.pass_context
@handle_error
@multi_user_do
def pwd(ctx):
    """显示当前工作目录"""

    pwd = _pwd(ctx)
    print(pwd)


# }}}


# File Operations
# {{{
@app.command()
@click.argument("remotepaths", nargs=-1, type=str)
@click.option("--desc", "-r", is_flag=True, help="逆序排列文件")
@click.option("--name", "-n", is_flag=True, help="依名字排序")
@click.option("--time", "-t", is_flag=True, help="依时间排序")
@click.option("--size", "-s", is_flag=True, help="依文件大小排序")
@click.option("--recursive", "-R", is_flag=True, help="递归列出文件")
@click.option("--include", "-I", type=str, help="筛选包含这个字符串的文件")
@click.option("--include-regex", "--IR", type=str, help="筛选包含这个正则表达式的文件")
@click.option("--exclude", "-E", type=str, help="筛选 不 包含这个字符串的文件")
@click.option(
    "--exclude-regex", "--ER", type=str, help="筛选 不 包含这个正则表达式的文件"
)
@click.option("--is-file", "-f", is_flag=True, help="筛选 非 目录文件")
@click.option("--is-dir", "-d", is_flag=True, help="筛选目录文件")
@click.option("--no-highlight", "--NH", is_flag=True, help="取消匹配高亮")
@click.option("--show-size", "-S", is_flag=True, help="显示文件大小")
@click.option("--show-date", "-D", is_flag=True, help="显示文件创建时间")
@click.option("--show-md5", "-M", is_flag=True, help="显示文件md5")
@click.option("--show-absolute-path", "-A", is_flag=True, help="显示文件绝对路径")
@click.option("--show-dl-link", "--DL", is_flag=True, help="显示文件下载连接")
@click.option("--show-hash-link", "--HL", is_flag=True, help="显示文件秒传连接")
@click.option(
    "--hash-link-protocol",
    "--HLP",
    type=click.Choice(PcsRapidUploadInfo.hash_link_protocols()),
    default=PcsRapidUploadInfo.default_hash_link_protocol(),
    help="显示文件hash链接，并指定协议",
)
@click.option(
    "--no-check-md5",
    "--NC",
    is_flag=True,
    help="显示文件 cs3l:// 连接时不检查md5",
)
@click.option(
    "--csv", is_flag=True, help="用 csv 格式显示，单行显示，推荐和 --DL 或 --HL 一起用"
)
@click.option("--only-dl-link", "--ODL", is_flag=True, help="只显示文件下载连接")
@click.option("--only-hash-link", "--OHL", is_flag=True, help="只显示文件秒传连接")
@click.pass_context
@handle_error
@multi_user_do
def ls(
    ctx,
    remotepaths,
    desc,
    name,
    time,
    size,
    recursive,
    include,
    include_regex,
    exclude,
    exclude_regex,
    is_file,
    is_dir,
    no_highlight,
    show_size,
    show_date,
    show_md5,
    show_absolute_path,
    show_dl_link,
    show_hash_link,
    hash_link_protocol,
    no_check_md5,
    csv,
    only_dl_link,
    only_hash_link,
):
    """列出网盘路径下的文件和对应的文件信息"""

    api = _recent_api(ctx)
    if not api:
        return

    sifters = []
    if include:
        sifters.append(IncludeSifter(include, regex=False))
    if include_regex:
        sifters.append(IncludeSifter(include_regex, regex=True))
    if exclude:
        sifters.append(ExcludeSifter(exclude, regex=False))
    if exclude_regex:
        sifters.append(ExcludeSifter(exclude_regex, regex=True))
    if is_file:
        sifters.append(IsFileSifter())
    if is_dir:
        sifters.append(IsDirSifter())

    pwd = _pwd(ctx)
    remotepaths = (join_path(pwd, r) for r in list(remotepaths) or (pwd,))

    rapiduploadinfo_file = _rapiduploadinfo_file(ctx)

    user_id, user_name = _recent_user_id_and_name(ctx)

    list_files(
        api,
        *remotepaths,
        desc=desc,
        name=name,
        time=time,
        size=size,
        recursive=recursive,
        sifters=sifters,
        highlight=not no_highlight,
        rapiduploadinfo_file=rapiduploadinfo_file,
        user_id=user_id,
        user_name=user_name,
        show_size=show_size,
        show_date=show_date,
        show_md5=show_md5,
        show_absolute_path=show_absolute_path,
        show_dl_link=show_dl_link,
        show_hash_link=show_hash_link,
        hash_link_protocol=hash_link_protocol,
        check_md5=not no_check_md5,
        csv=csv,
        only_dl_link=only_dl_link,
        only_hash_link=only_hash_link,
    )


@app.command()
@click.argument("remotepaths", nargs=-1, type=str)
@click.option("--recursive", "-R", is_flag=True, help="递归计算所有文件")
@click.option("--include", "-I", type=str, help="筛选包含这个字符串的文件")
@click.option("--include-regex", "--IR", type=str, help="筛选包含这个正则表达式的文件")
@click.option("--exclude", "-E", type=str, help="筛选 不 包含这个字符串的文件")
@click.option(
    "--exclude-regex", "--ER", type=str, help="筛选 不 包含这个正则表达式的文件"
)
@click.pass_context
@handle_error
@multi_user_do
def du(ctx, remotepaths, recursive, include, include_regex, exclude, exclude_regex):
    """统计网盘路径下的文件所占用的空间"""

    api = _recent_api(ctx)
    if not api:
        return

    sifters = []
    if include:
        sifters.append(IncludeSifter(include, regex=False))
    if include_regex:
        sifters.append(IncludeSifter(include_regex, regex=True))
    if exclude:
        sifters.append(ExcludeSifter(exclude, regex=False))
    if exclude_regex:
        sifters.append(ExcludeSifter(exclude_regex, regex=True))

    pwd = _pwd(ctx)
    remotepaths = (join_path(pwd, r) for r in list(remotepaths) or (pwd,))

    disk_usage(
        api,
        *remotepaths,
        recursive=recursive,
        sifters=sifters,
    )


@app.command()
@click.argument("keyword", nargs=1, type=str)
@click.argument("remotedir", nargs=1, type=str, default="")
@click.option("--recursive", "-R", is_flag=True, help="递归搜索文件")
@click.option("--include", "-I", type=str, help="筛选包含这个字符串的文件")
@click.option("--include-regex", "--IR", type=str, help="筛选包含这个正则表达式的文件")
@click.option("--exclude", "-E", type=str, help="筛选 不 包含这个字符串的文件")
@click.option(
    "--exclude-regex", "--ER", type=str, help="筛选 不 包含这个正则表达式的文件"
)
@click.option("--is-file", "-f", is_flag=True, help="筛选 非 目录文件")
@click.option("--is-dir", "-d", is_flag=True, help="筛选目录文件")
@click.option("--no-highlight", "--NH", is_flag=True, help="取消匹配高亮")
@click.option("--show-size", "-S", is_flag=True, help="显示文件大小")
@click.option("--show-date", "-D", is_flag=True, help="显示文件创建时间")
@click.option("--show-md5", "-M", is_flag=True, help="显示文件md5")
@click.option("--csv", is_flag=True, help="用 csv 格式显示")
@click.pass_context
@handle_error
@multi_user_do
def search(
    ctx,
    keyword,
    remotedir,
    recursive,
    include,
    include_regex,
    exclude,
    exclude_regex,
    is_file,
    is_dir,
    no_highlight,
    show_size,
    show_date,
    show_md5,
    csv,
):
    """搜索包含 `keyword` 的文件"""

    api = _recent_api(ctx)
    if not api:
        return

    sifters = []
    if include:
        sifters.append(IncludeSifter(include, regex=False))
    if include_regex:
        sifters.append(IncludeSifter(include_regex, regex=True))
    if exclude:
        sifters.append(ExcludeSifter(exclude, regex=False))
    if exclude_regex:
        sifters.append(ExcludeSifter(exclude_regex, regex=True))
    if is_file:
        sifters.append(IsFileSifter())
    if is_dir:
        sifters.append(IsDirSifter())

    pwd = _pwd(ctx)
    remotedir = join_path(pwd, remotedir)

    _search(
        api,
        keyword,
        remotedir,
        recursive=recursive,
        sifters=sifters,
        highlight=not no_highlight,
        show_size=show_size,
        show_date=show_date,
        show_md5=show_md5,
        csv=csv,
    )


@app.command()
@click.argument("remotepath", nargs=1, type=str)
@click.option("--encoding", "-e", type=str, help="文件编码，默认自动解码")
@click.option("--no-decrypt", "--ND", is_flag=True, help="不解密")
@click.option(
    "--encrypt-password",
    "--ep",
    type=str,
    default=None,
    help="加密密码，默认使用用户设置的",
)
@click.pass_context
@handle_error
@multi_user_do
def cat(ctx, remotepath, encoding, no_decrypt, encrypt_password):
    """显示文件内容"""

    api = _recent_api(ctx)
    if not api:
        return

    pwd = _pwd(ctx)
    remotepath = join_path(pwd, remotepath)

    if no_decrypt:
        encrypt_password = b""
    else:
        encrypt_password = encrypt_password or _encrypt_password(ctx)

    _cat(api, remotepath, encoding=encoding, encrypt_password=encrypt_password)


@app.command()
@click.argument("remotedirs", nargs=-1, type=str)
@click.option("--show", "-S", is_flag=True, help="显示目录")
@click.pass_context
@handle_error
@multi_user_do
def mkdir(ctx, remotedirs, show):
    """创建目录"""

    api = _recent_api(ctx)
    if not api:
        return

    pwd = _pwd(ctx)
    remotedirs = (join_path(pwd, d) for d in remotedirs)

    file_operators.makedir(api, *remotedirs, show=show)


@app.command()
@click.argument("remotepaths", nargs=-1, type=str)
@click.option("--show", "-S", is_flag=True, help="显示结果")
@click.pass_context
@handle_error
@multi_user_do
def move(ctx, remotepaths, show):
    """移动文件

    \b
    examples:
        move /file1 /file2 /to/dir
    """

    api = _recent_api(ctx)
    if not api:
        return

    pwd = _pwd(ctx)
    remotepaths = [join_path(pwd, r) for r in remotepaths]

    if len(remotepaths) < 2:
        ctx.fail("remote paths < 2")
    file_operators.move(api, *remotepaths, show=show)


@app.command()
@click.argument("source", nargs=1, type=str)
@click.argument("dest", nargs=1, type=str)
@click.option("--show", "-S", is_flag=True, help="显示结果")
@click.pass_context
@handle_error
@multi_user_do
def rename(ctx, source, dest, show):
    """文件重命名

    \b
    examples:
        rename /path/to/far /to/here/foo
    """

    api = _recent_api(ctx)
    if not api:
        return

    pwd = _pwd(ctx)
    source = join_path(pwd, source)
    dest = join_path(pwd, dest)

    file_operators.rename(api, source, dest, show=show)


@app.command()
@click.argument("remotepaths", nargs=-1, type=str)
@click.option("--show", "-S", is_flag=True, help="显示结果")
@click.pass_context
@handle_error
@multi_user_do
def copy(ctx, remotepaths, show):
    """拷贝文件

    \b
    examples:
        copy /file1 /file2 /to/dir
    """

    api = _recent_api(ctx)
    if not api:
        return

    pwd = _pwd(ctx)
    remotepaths = [join_path(pwd, r) for r in remotepaths]

    if len(remotepaths) < 2:
        ctx.fail("remote paths < 2")
    file_operators.copy(api, *remotepaths, show=show)


@app.command()
@click.argument("remotepaths", nargs=-1, type=str)
@click.pass_context
@handle_error
@multi_user_do
def remove(ctx, remotepaths):
    """删除文件"""

    api = _recent_api(ctx)
    if not api:
        return

    pwd = _pwd(ctx)
    remotepaths = (join_path(pwd, r) for r in remotepaths)

    file_operators.remove(api, *remotepaths)


@app.command()
@click.argument("remotepaths", nargs=-1, type=str)
@click.option(
    "--outdir",
    "-o",
    nargs=1,
    type=str,
    default=".",
    help="指定下载本地目录，默认为当前目录",
)
@click.option("--recursive", "-R", is_flag=True, help="递归下载")
@click.option(
    "--from-index",
    "-f",
    type=int,
    default=0,
    help="从所有目录中的第几个文件开始下载，默认为0（第一个）",
)
@click.option("--include", "-I", type=str, help="筛选包含这个字符串的文件")
@click.option("--include-regex", "--IR", type=str, help="筛选包含这个正则表达式的文件")
@click.option("--exclude", "-E", type=str, help="筛选 不 包含这个字符串的文件")
@click.option(
    "--exclude-regex", "--ER", type=str, help="筛选 不 包含这个正则表达式的文件"
)
@click.option(
    "-d",
    "--downloader",
    type=click.Choice([d.name for d in Downloader]),
    default=DEFAULT_DOWNLOADER.name,
    help="""指定下载应用

    \b
    默认为 me (BaiduPCS-Py 自己的下载器，支持断续下载)
        me 使用多文件并发下载。

    除 me 外，其他下载器，不使用多文件并发下载，使用一个文件多链接下载。
    如果需要下载多个小文件推荐使用 me，如果需要下载少量大文件推荐使用其他下载器。

    \b
    aget_py (https://github.com/PeterDing/aget) 默认安装
    aget_rs (下载 https://github.com/PeterDing/aget-rs/releases)
    aria2 (下载 https://github.com/aria2/aria2/releases)
    """,
)
@click.option(
    "--downloader-params", "--DP", multiple=True, type=str, help="第三方下载器参数"
)
@click.option(
    "--concurrency",
    "-s",
    type=int,
    default=DEFAULT_CONCURRENCY,
    help="下载同步链接数，默认为5。数子越大下载速度越快，但是容易被百度封锁",
)
@click.option(
    "--chunk-size", "-k", type=str, default=DEFAULT_CHUNK_SIZE, help="同步链接分块大小"
)
@click.option("--no-decrypt", "--ND", is_flag=True, help="不解密")
@click.option("--quiet", "-q", is_flag=True, help="取消第三方下载应用输出")
@click.option("--out-cmd", "--OC", is_flag=True, help="输出第三方下载应用命令")
@click.option(
    "--encrypt-password",
    "--ep",
    type=str,
    default=None,
    help="加密密码，默认使用用户设置的",
)
@click.pass_context
@handle_error
@multi_user_do
def download(
    ctx,
    remotepaths,
    outdir,
    recursive,
    from_index,
    include,
    include_regex,
    exclude,
    exclude_regex,
    downloader,
    downloader_params,
    concurrency,
    chunk_size,
    no_decrypt,
    quiet,
    out_cmd,
    encrypt_password,
):
    """下载文件"""

    if out_cmd:
        assert downloader != Downloader.me.name, "输出命令只能用于第三方下载应用"

    api = _recent_api(ctx)
    if not api:
        return

    sifters = []
    if include:
        sifters.append(IncludeSifter(include, regex=False))
    if include_regex:
        sifters.append(IncludeSifter(include_regex, regex=True))
    if exclude:
        sifters.append(ExcludeSifter(exclude, regex=False))
    if exclude_regex:
        sifters.append(ExcludeSifter(exclude_regex, regex=True))

    pwd = _pwd(ctx)
    remotepaths = [join_path(pwd, r) for r in remotepaths]

    if no_decrypt:
        encrypt_password = b""
    else:
        encrypt_password = encrypt_password or _encrypt_password(ctx)

    if not quiet and downloader == "me":
        init_progress_bar()

    _download(
        api,
        remotepaths,
        outdir,
        sifters=sifters,
        recursive=recursive,
        from_index=from_index,
        downloader=getattr(Downloader, downloader),
        downloadparams=DownloadParams(
            concurrency=concurrency,
            chunk_size=chunk_size,
            quiet=quiet,
            downloader_params=downloader_params,
        ),
        out_cmd=out_cmd,
        encrypt_password=encrypt_password,
    )


@app.command()
@click.argument("remotepaths", nargs=-1, type=str)
@click.option("--recursive", "-R", is_flag=True, help="递归播放")
@click.option(
    "--from-index",
    "-f",
    type=int,
    default=0,
    help="从所有目录中的第几个文件开始播放，默认为0（第一个）",
)
@click.option("--include", "-I", type=str, help="筛选包含这个字符串的文件")
@click.option("--include-regex", "--IR", type=str, help="筛选包含这个正则表达式的文件")
@click.option("--exclude", "-E", type=str, help="筛选 不 包含这个字符串的文件")
@click.option(
    "--exclude-regex", "--ER", type=str, help="筛选 不 包含这个正则表达式的文件"
)
@click.option(
    "-p",
    "--player",
    type=click.Choice([d.name for d in Player]),
    default=DEFAULT_PLAYER.name,
    help="""指定第三方播放器

    \b
    默认为 mpv (https://mpv.io),
    """,
)
@click.option(
    "--player-params", "--PP", multiple=True, type=str, help="第三方播放器参数"
)
@click.option("--m3u8", "-m", is_flag=True, help="获取m3u8文件并播放")
@click.option("--quiet", "-q", is_flag=True, help="取消第三方播放器输出")
@click.option("--shuffle", "--sf", is_flag=True, help="随机播放")
@click.option(
    "--ignore-ext", "--IE", is_flag=True, help="不用文件名后缀名来判断媒体文件"
)
@click.option("--out-cmd", "--OC", is_flag=True, help="输出第三方播放器命令")
@click.option(
    "--use-local-server",
    "-s",
    is_flag=True,
    help="使用本地服务器播放。大于100MB的媒体文件无法直接播放，需要使用本地服务器播放",
)
@click.option(
    "--encrypt-password",
    "--ep",
    type=str,
    default=None,
    help="加密密码，默认使用用户设置的",
)
@click.pass_context
@handle_error
@multi_user_do
def play(
    ctx,
    remotepaths,
    recursive,
    from_index,
    include,
    include_regex,
    exclude,
    exclude_regex,
    player,
    player_params,
    m3u8,
    quiet,
    shuffle,
    ignore_ext,
    out_cmd,
    use_local_server,
    encrypt_password,
):
    """播放媒体文件"""

    api = _recent_api(ctx)
    if not api:
        return

    sifters = []
    if include:
        sifters.append(IncludeSifter(include, regex=False))
    if include_regex:
        sifters.append(IncludeSifter(include_regex, regex=True))
    if exclude:
        sifters.append(ExcludeSifter(exclude, regex=False))
    if exclude_regex:
        sifters.append(ExcludeSifter(exclude_regex, regex=True))

    pwd = _pwd(ctx)
    remotepaths = [join_path(pwd, r) for r in remotepaths]

    local_server = ""
    if use_local_server:
        encrypt_password = encrypt_password or _encrypt_password(ctx)

        host = "localhost"
        port = random_avail_port()

        local_server = f"http://{host}:{port}"

        ps = Process(
            target=start_server,
            args=(
                api,
                "/",
            ),
            kwargs=dict(
                host=host,
                port=port,
                workers=CPU_NUM,
                encrypt_password=encrypt_password,
                log_level="warning",
            ),
        )
        ps.start()
        time.sleep(1)

    _play(
        api,
        remotepaths,
        sifters=sifters,
        recursive=recursive,
        from_index=from_index,
        player=getattr(Player, player),
        player_params=player_params,
        m3u8=m3u8,
        quiet=quiet,
        shuffle=shuffle,
        ignore_ext=ignore_ext,
        out_cmd=out_cmd,
        local_server=local_server,
    )

    if use_local_server:
        ps.terminate()


@app.command()
@click.argument("localpaths", nargs=-1, type=str)
@click.argument("remotedir", nargs=1, type=str)
@click.option(
    "--upload-type",
    "-t",
    type=click.Choice([t.name for t in UploadType]),
    default=UploadType.Many.name,
    help="上传方式，Many: 同时上传多个文件，One: 一次只上传一个文件，但同时上传文件的多个分片",
)
@click.option(
    "--encrypt-password",
    "--ep",
    type=str,
    default=None,
    help="加密密码，默认使用用户设置的",
)
@click.option(
    "--encrypt-type",
    "-e",
    type=click.Choice([t.name for t in EncryptType]),
    default=EncryptType.No.name,
    help="文件加密方法，默认为 No 不加密",
)
@click.option(
    "--max-workers",
    "-w",
    type=int,
    default=CPU_NUM,
    help="同时上传连接数量，默认为 CPU 核数",
)
@click.option("--no-ignore-existing", "--NI", is_flag=True, help="上传已经存在的文件")
@click.option("--no-show-progress", "--NP", is_flag=True, help="不显示上传进度")
@click.option(
    "--check-md5",
    "--CM",
    is_flag=True,
    help="分段上传后检查md5。注意检查上传后大文件的md5，可能会花数分中（2G 的文件需要大约5分钟）",
)
@click.pass_context
@handle_error
@multi_user_do
def upload(
    ctx,
    localpaths,
    remotedir,
    upload_type,
    encrypt_password,
    encrypt_type,
    max_workers,
    no_ignore_existing,
    no_show_progress,
    check_md5,
):
    """上传文件"""

    # Keyboard listener start
    keyboard_listener_start()

    api = _recent_api(ctx)
    if not api:
        return

    encrypt_password = encrypt_password or _encrypt_password(ctx)
    if encrypt_type != EncryptType.No.name and not encrypt_password:
        raise ValueError(f"Encrypting with {encrypt_type} must have a key")

    pwd = _pwd(ctx)
    remotedir = join_path(pwd, remotedir)

    rapiduploadinfo_file = _rapiduploadinfo_file(ctx)

    user_id, user_name = _recent_user_id_and_name(ctx)

    from_to_list = from_tos(localpaths, remotedir)

    if not no_show_progress:
        init_progress_bar()

    _upload(
        api,
        from_to_list,
        upload_type=getattr(UploadType, upload_type),
        encrypt_password=encrypt_password,
        encrypt_type=getattr(EncryptType, encrypt_type),
        max_workers=max_workers,
        ignore_existing=not no_ignore_existing,
        show_progress=not no_show_progress,
        rapiduploadinfo_file=rapiduploadinfo_file,
        user_id=user_id,
        user_name=user_name,
        check_md5=check_md5,
    )


@app.command()
@click.argument("localdir", nargs=1, type=str)
@click.argument("remotedir", nargs=1, type=str)
@click.option(
    "--encrypt-password",
    "--ep",
    type=str,
    default=None,
    help="加密密码，默认使用用户设置的",
)
@click.option(
    "--encrypt-type",
    "-e",
    type=click.Choice([t.name for t in EncryptType]),
    default=EncryptType.No.name,
    help="文件加密方法，默认为 No 不加密",
)
@click.option("--max-workers", "-w", type=int, default=CPU_NUM, help="同时上传文件数")
@click.option("--no-show-progress", "--NP", is_flag=True, help="不显示上传进度")
@click.option(
    "--check-md5",
    "--CM",
    is_flag=True,
    help="分段上传后检查md5。注意检查上传后大文件的md5，可能会花数分中（2G 的文件需要大约5分钟）",
)
@click.pass_context
@handle_error
@multi_user_do
def sync(
    ctx,
    localdir,
    remotedir,
    encrypt_password,
    encrypt_type,
    max_workers,
    no_show_progress,
    check_md5,
):
    """同步本地目录到远端"""

    # Keyboard listener start
    keyboard_listener_start()

    api = _recent_api(ctx)
    if not api:
        return

    pwd = _pwd(ctx)
    remotedir = join_path(pwd, remotedir)

    encrypt_password = encrypt_password or _encrypt_password(ctx)
    if encrypt_type != EncryptType.No.name and not encrypt_password:
        raise ValueError(f"Encrypting with {encrypt_type} must have a key")

    rapiduploadinfo_file = _rapiduploadinfo_file(ctx)

    user_id, user_name = _recent_user_id_and_name(ctx)

    _sync(
        api,
        localdir,
        remotedir,
        encrypt_password=encrypt_password,
        encrypt_type=getattr(EncryptType, encrypt_type),
        max_workers=max_workers,
        show_progress=not no_show_progress,
        rapiduploadinfo_file=rapiduploadinfo_file,
        user_id=user_id,
        user_name=user_name,
        check_md5=check_md5,
    )


# }}}


# Rapid Upload
# {{{
@app.command()
@click.argument("ids", nargs=-1, type=int, default=None)
@click.option("--filename", "-f", is_flag=True, help="按文件名排序")
@click.option("--time", "-t", is_flag=True, help="按时间排序")
@click.option("--size", "-s", is_flag=True, help="按文件大小排序")
@click.option("--localpath", "-l", is_flag=True, help="按本地名排序")
@click.option("--remotepath", "-r", is_flag=True, help="按远端名排序")
@click.option("--userid", "-u", is_flag=True, help="按用户id排序")
@click.option("--username", "-n", is_flag=True, help="按用户名排序")
@click.option("--desc", "-d", is_flag=True, help="按逆序排序")
@click.option("--limit", "-L", type=int, default=-1, help="限制列出文件个数")
@click.option("--offset", "-O", type=int, default=-1, help="列出偏移位")
@click.option(
    "--hash-link-protocol",
    "--HLP",
    type=click.Choice(PcsRapidUploadInfo.hash_link_protocols()),
    default=PcsRapidUploadInfo.default_hash_link_protocol(),
    help="hash link 协议，默认 cs3l",
)
@click.option("--show-all", "-A", is_flag=True, help="显示文件所有信息")
@click.option("--only-hash-link", "--OHL", is_flag=True, help="只显示文件秒传连接")
@click.pass_context
@handle_error
def rplist(
    ctx,
    ids,
    filename,
    time,
    size,
    localpath,
    remotepath,
    userid,
    username,
    desc,
    limit,
    offset,
    hash_link_protocol,
    show_all,
    only_hash_link,
):
    """列出保存的文件秒传信息"""

    rapiduploadinfo_file = _rapiduploadinfo_file(ctx)
    rapid_upload_list(
        rapiduploadinfo_file,
        ids=ids or [],
        filename=filename,
        time=time,
        size=size,
        localpath=localpath,
        remotepath=remotepath,
        user_id=userid,
        user_name=username,
        desc=desc,
        limit=limit,
        offset=offset,
        hash_link_protocol=hash_link_protocol,
        show_all=show_all,
        only_hash_link=only_hash_link,
    )


@app.command()
@click.argument("keyword", nargs=1, type=str)
@click.option("--filename", "--fn", is_flag=True, help="在文件名中搜索")
@click.option("--localpath", "--lp", is_flag=True, help="在本地路径中搜索")
@click.option("--remotepath", "--rp", is_flag=True, help="在远端路径中搜索")
@click.option("--username", "--un", is_flag=True, help="在用户名中搜索")
@click.option(
    "--md5", "-m", is_flag=True, help="在md5中搜索；注意保存的文件md5都是小写字符"
)
@click.option(
    "--hash-link-protocol",
    "--HLP",
    type=click.Choice(PcsRapidUploadInfo.hash_link_protocols()),
    default=PcsRapidUploadInfo.default_hash_link_protocol(),
    help="hash link 协议，默认 cs3l",
)
@click.option("--show-all", "-A", is_flag=True, help="显示文件所有信息")
@click.option("--only-hash-link", "--OHL", is_flag=True, help="只显示文件秒传连接")
@click.pass_context
@handle_error
def rpsearch(
    ctx,
    keyword,
    filename,
    localpath,
    remotepath,
    username,
    md5,
    hash_link_protocol,
    show_all,
    only_hash_link,
):
    """搜索保存的文件秒传信息"""

    rapiduploadinfo_file = _rapiduploadinfo_file(ctx)
    rapid_upload_search(
        rapiduploadinfo_file,
        keyword,
        in_filename=filename,
        in_localpath=localpath,
        in_remotepath=remotepath,
        in_user_name=username,
        in_md5=md5,
        hash_link_protocol=hash_link_protocol,
        show_all=show_all,
        only_hash_link=only_hash_link,
    )


@app.command()
@click.argument("ids", nargs=-1, type=int)
@click.pass_context
@handle_error
def rpdelete(ctx, ids):
    """按数据库id删除保存的文件秒传信息"""

    rapiduploadinfo_file = _rapiduploadinfo_file(ctx)
    rapid_upload_delete(rapiduploadinfo_file, ids)


@app.command()
@click.argument("remotedir", nargs=1, default="", type=str)
@click.option("--link", "-l", type=str, help="秒传连接")
@click.option(
    "--input-file",
    "-i",
    type=str,
    help="从指定文件获取要使用的秒传连接；只能是一行一个秒传连接",
)
@click.option("--slice-md5", "--sm", type=str, help="文件前 256KB md5")
@click.option("--content-md5", "--cm", type=str, help="文件 md5")
@click.option("--content-crc32", "--cc", type=int, help="文件 crc32, 可以为空")
@click.option("--content-length", "--cl", type=int, help="文件长度")
@click.option(
    "--filename",
    "--fn",
    type=str,
    help="文件名，如果这里设置了，将会覆盖 link 中的文件名",
)
@click.option(
    "--no-ignore-existing", "--NI", is_flag=True, help="上传且覆盖已经存在的文件"
)
@click.option("--max-workers", "-w", type=int, default=CPU_NUM, help="同时上传文件数")
@click.pass_context
@handle_error
@multi_user_do
def rp(
    ctx,
    remotedir,
    link,
    input_file,
    slice_md5,
    content_md5,
    content_crc32,
    content_length,
    filename,
    no_ignore_existing,
    max_workers,
):
    """用秒传连接或参数上传

    如果设置了 --input-file，将会忽略 `--link` 和 指定的特征参数。
    如果设置了 --link，那么其他参数都不需要了。
    如果设置了 --filename，将会覆盖 link 中的文件名。
    """

    api = _recent_api(ctx)
    if not api:
        return

    assert (
        link or input_file or all([slice_md5, content_md5, content_length, filename])
    ), "No params"

    pwd = _pwd(ctx)
    remotedir = join_path(pwd, remotedir)

    rapiduploadinfo_file = _rapiduploadinfo_file(ctx)

    user_id, user_name = _recent_user_id_and_name(ctx)

    if input_file and os.path.exists(input_file):
        lines = open(input_file).readlines()
        links = [line.strip() for line in lines if line.strip()]
        rapid_upload_links(
            api,
            remotedir,
            links=links,
            no_ignore_existing=no_ignore_existing,
            rapiduploadinfo_file=rapiduploadinfo_file,
            user_id=user_id,
            user_name=user_name,
            max_workers=max_workers,
        )
    else:
        rapid_upload(
            api,
            remotedir,
            link=link,
            slice_md5=slice_md5,
            content_md5=content_md5,
            content_crc32=content_crc32,
            content_length=content_length,
            filename=filename,
            no_ignore_existing=no_ignore_existing,
            rapiduploadinfo_file=rapiduploadinfo_file,
            user_id=user_id,
            user_name=user_name,
        )


# }}}


# Share
# {{{
@app.command()
@click.argument("remotepaths", nargs=-1, type=str)
@click.option("--password", "-p", type=str, help="设置秘密，4个字符。默认没有秘密")
@click.option("--period-time", "--pt", type=int, help="设置分享有效期，单位为天")
@click.pass_context
@handle_error
@multi_user_do
def share(ctx, remotepaths, password, period_time):
    """分享文件

    \b
    examples:
        share /path1 path2
    """
    assert not password or len(password) == 4, "`password` must be 4 letters"

    api = _recent_api(ctx)
    if not api:
        return

    pwd = _pwd(ctx)
    remotepaths = (join_path(pwd, r) for r in remotepaths)

    _share.share_files(api, *remotepaths, password=password, period=period_time or 0)


@app.command()
@click.option(
    "--show-all",
    "-A",
    is_flag=True,
    help="显示所有分享的链接，默认只显示有效的分享链接",
)
@click.pass_context
@handle_error
@multi_user_do
def shared(ctx, show_all):
    """列出分享链接"""

    api = _recent_api(ctx)
    if not api:
        return

    _share.list_shared(api, show_all=show_all)


@app.command()
@click.argument("share_ids", nargs=-1, type=int)
@click.pass_context
@handle_error
@multi_user_do
def cancelshared(ctx, share_ids):
    """取消分享链接"""

    api = _recent_api(ctx)
    if not api:
        return

    _share.cancel_shared(api, *share_ids)


@app.command()
@click.argument("shared_url", nargs=1, type=str)
@click.option("--password", "-p", type=str, help="链接密码，如果没有不用设置")
@click.option("--no-show-vcode", "--NV", is_flag=True, help="不显示验证码")
@click.pass_context
@handle_error
@multi_user_do
def listsharedpaths(ctx, shared_url, password, no_show_vcode):
    """列出其他用户分享链接中的文件"""

    assert not password or len(password) == 4, "`password` must be 4 letters"

    api = _recent_api(ctx)
    if not api:
        return

    _share.list_shared_paths(
        api,
        shared_url,
        password=password,
        show_vcode=not no_show_vcode,
    )


@app.command()
@click.argument("shared_url", nargs=1, type=str)
@click.argument("remotedir", nargs=1, type=str)
@click.option("--password", "-p", type=str, help="链接密码，如果没有不用设置")
@click.option("--no-show-vcode", "--NV", is_flag=True, help="不显示验证码")
@click.pass_context
@handle_error
@multi_user_do
def save(ctx, shared_url, remotedir, password, no_show_vcode):
    """保存其他用户分享的链接"""

    assert not password or len(password) == 4, "`password` must be 4 letters"

    api = _recent_api(ctx)
    if not api:
        return

    pwd = _pwd(ctx)
    remotedir = join_path(pwd, remotedir)

    _share.save_shared(
        api,
        shared_url,
        remotedir,
        password=password,
        show_vcode=not no_show_vcode,
    )


# }}}


# Cloud
# {{{
@app.command()
@click.argument("task_urls", nargs=-1, type=str)
@click.argument("remotedir", nargs=1, type=str)
@click.option(
    "--file-type",
    "-t",
    type=str,
    default="m",
    help="要下载的文件类型，m:媒体文件，i:图片文件，d:文档文件，c:压缩文件，a:所有文件。用','分割。",
)
@click.pass_context
@handle_error
@multi_user_do
def add(ctx, task_urls, remotedir, file_type):
    """添加离线下载任务

    \b
    如果添加 magnet 连接需要指定要下载的文件类型：
    m:媒体文件 (默认)
    i:图片文件
    d:文档文件
    c:压缩文件
    a:所有文件

    比如要下载所有的媒体文件和文档文件：
    --file-type 'm,d'
    """

    api = _recent_api(ctx)
    if not api:
        return

    pwd = _pwd(ctx)
    remotedir = join_path(pwd, remotedir)

    file_types = [_cloud.FileType.from_(addr) for addr in file_type.split(",")]

    for url in task_urls:
        _cloud.add_task(api, url, remotedir, file_types=file_types)


@app.command()
@click.argument("task_ids", nargs=-1, type=str)
@click.pass_context
@handle_error
@multi_user_do
def tasks(ctx, task_ids):
    """列出离线下载任务。也可列出给定id的任务"""

    api = _recent_api(ctx)
    if not api:
        return

    if not task_ids:
        _cloud.list_tasks(api)
    else:
        _cloud.tasks(api, *task_ids)


@app.command()
@click.pass_context
@handle_error
@multi_user_do
def cleartasks(ctx):
    """清除已经下载完和下载失败的任务"""

    api = _recent_api(ctx)
    if not api:
        return

    _cloud.clear_tasks(api)


@app.command()
@click.argument("task_ids", nargs=-1, type=str)
@click.pass_context
@handle_error
@multi_user_do
def canceltasks(ctx, task_ids):
    """取消下载任务"""

    api = _recent_api(ctx)
    if not api:
        return

    for task_id in task_ids:
        _cloud.cancel_task(api, task_id)


@app.command()
@click.option("--yes", is_flag=True, help="确认并直接运行")
@click.pass_context
@handle_error
@multi_user_do
def purgetasks(ctx, yes):
    """删除所有离线下载任务"""

    api = _recent_api(ctx)
    if not api:
        return

    if not yes:
        if not Confirm.ask("确定删除[i red]所有的[/i red]离线下载任务?", default=False):
            return
    _cloud.purge_all_tasks(api)


# }}}


# Server
# {{{
@app.command()
@click.argument("root_dir", type=str, default="/", required=False)
@click.option("--path", type=str, default="/", help="服务路径，默认为 “/”")
@click.option("--host", "-h", type=str, default="localhost", help="监听 host")
@click.option("--port", "-p", type=int, default=8000, help="监听 port")
@click.option("--workers", "-w", type=int, default=CPU_NUM, help="进程数")
@click.option(
    "--encrypt-password",
    "--ep",
    type=str,
    default=None,
    help="加密密码，默认使用用户设置的",
)
@click.option("--username", type=str, default=None, help="HTTP Basic Auth 用户名")
@click.option("--password", type=str, default=None, help="HTTP Basic Auth 密钥")
@click.pass_context
@handle_error
@multi_user_do
def server(
    ctx, root_dir, path, host, port, workers, encrypt_password, username, password
):
    """开启 HTTP 服务"""

    api = _recent_api(ctx)
    if not api:
        return

    encrypt_password = encrypt_password or _encrypt_password(ctx)

    if username:
        assert password, "Must set password"

    start_server(
        api,
        root_dir=root_dir,
        path=path,
        host=host,
        port=port,
        workers=workers,
        encrypt_password=encrypt_password,
        username=username,
        password=password,
    )


# }}}



import hashlib
import logging
import os
import time
from enum import Enum
from io import BytesIO, UnsupportedOperation
from pathlib import Path
from random import Random
from typing import (
    IO,
    Any,
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    Tuple,
    Union,
    cast,
)
from zlib import crc32

import requests  # type: ignore
from requests import Response

from . import constant
from .crypto import (
    AES256CBCCryptography,
    ChaCha20Cryptography,
    Cryptography,
    SimpleCryptography,
    aes256cbc_decrypt,
    aes256cbc_encrypt,
    calu_crc32_and_md5,
    calu_md5,
    generate_key_iv,
    generate_salt,
    padding_key,
    padding_size,
    pkcs7_padding,
    pkcs7_unpadding,
    random_bytes,
    random_sys_bytes,
)
from .log import LogLevels, TLogLevel, get_logger
from .number import u8x8_to_u64, u64_to_u8x8

_LOG_LEVEL = os.getenv("LOG_LEVEL", "CRITICAL").upper()
if _LOG_LEVEL not in LogLevels:
    _LOG_LEVEL = "CRITICAL"
logger = get_logger(__name__, level=cast(TLogLevel, _LOG_LEVEL))

READ_SIZE = 65535

# This is the threshold of range request setted by Baidu server
MAX_CHUNK_SIZE = 50 * constant.OneM

BAIDUPCS_PY_CRYPTO_MAGIC_CODE = b"\x00@@#__BAIDUPCS_PY__CRYPTO__#@@\x00\xff"
ENCRYPT_HEAD_LEN = len(BAIDUPCS_PY_CRYPTO_MAGIC_CODE) + 1 + 16 + 8
PADDED_ENCRYPT_HEAD_LEN = padding_size(ENCRYPT_HEAD_LEN, 16)
PADDED_ENCRYPT_HEAD_WITH_SALT_LEN = PADDED_ENCRYPT_HEAD_LEN + 8


def total_len(o: Any) -> int:
    """Read obj len"""

    if hasattr(o, "__len__"):
        return len(o)

    if hasattr(o, "seekable"):
        if o.seekable():
            _offset = o.tell()
            _len = o.seek(0, 2)
            o.seek(_offset, 0)
            return _len

    if hasattr(o, "len"):
        return o.len

    if hasattr(o, "fileno"):
        try:
            fileno = o.fileno()
        except UnsupportedOperation:
            pass
        else:
            return os.fstat(fileno).st_size

    if hasattr(o, "getvalue"):
        # e.g. BytesIO, cStringIO.StringIO
        return len(o.getvalue())

    raise TypeError("Unsupported Operation: no method to get len")


class ChunkIO(IO):
    """Wrap a IO, only read `size` bytes"""

    def __init__(self, io: IO, size: int):
        self._io = io
        self._size = size
        self._offset = 0
        self._io_offset = io.tell()

    def __len__(self) -> int:
        return self._size

    def read(self, size: int = -1) -> Optional[bytes]:
        remain = self._size - self._offset
        if size == -1:
            size = remain
        else:
            size = min(size, remain)

        data = self._io.read(size) or b""
        if data:
            self._offset += len(data)
        return data

    def seek(self, offset: int, whence: int = 0) -> int:
        if not self.seekable():
            raise ValueError(f"{self._io.__class__.__name__} is not support seeking")

        if whence == 0:
            if offset < 0:
                raise ValueError(f"Negative seek position {offset}")
            pass
        elif whence == 1:
            offset += self._offset
        elif whence == 2:
            offset = self._size - offset

        offset = max(0, offset)
        offset = min(self._size, offset)
        self._io.seek(self._io_offset + offset)
        self._offset = offset
        return self._offset

    def tell(self) -> int:
        return self._offset

    def seekable(self) -> bool:
        return self._io.seekable()

    def writable(self):
        return False

    def close(self):
        pass


def sample_data(io: IO, rg: Random, size: int) -> bytes:
    """Sample data with size"""

    _len = io.seek(0, 2)
    data = b""
    for _ in range(size):
        i = rg.randint(0, max(0, _len - 1))
        io.seek(i, 0)
        data += io.read(1)
    io.seek(0, 0)
    return data


def rapid_upload_params(io: IO) -> Tuple[str, str, int, int]:
    io_len = 0
    chunk_size = constant.OneM
    crc32_v = 0
    md5_v = hashlib.md5()

    buf = io.read(256 * constant.OneK)
    io_len += len(buf)
    slice_md5 = calu_md5(buf)
    md5_v.update(buf)
    t = io.read(chunk_size - 256 * constant.OneK)
    io_len += len(t)
    md5_v.update(t)
    buf += t
    crc32_v = crc32(buf, crc32_v).conjugate()

    # content_crc32, content_md5 = calu_crc32_and_md5(io, constant.OneM)

    while True:
        buf = io.read(chunk_size)
        if buf:
            md5_v.update(buf)
            crc32_v = crc32(buf, crc32_v).conjugate()
            io_len += len(buf)
        else:
            break

    content_crc32, content_md5 = crc32_v.conjugate() & 0xFFFFFFFF, md5_v.hexdigest()

    # if isinstance(io, EncryptIO):
    #     io_len = len(io)
    # else:
    #     io_len = io.seek(0, 2)
    #     io.seek(0, 0)

    return slice_md5, content_md5, content_crc32, io_len


def rapid_upload_params2(localPath: Path) -> Tuple[str, str, int, int]:
    buf = localPath.open("rb").read(constant.OneM)
    slice_md5 = calu_md5(buf[: 256 * constant.OneK])
    content_crc32, content_md5 = calu_crc32_and_md5(localPath.open("rb"), constant.OneM)
    return slice_md5, content_md5, content_crc32, localPath.stat().st_size


def generate_nonce_or_iv(salt: bytes, io: IO) -> bytes:
    io_len = total_len(io)
    rg = Random(salt)
    sample = sample_data(io, rg, 16)
    return random_bytes(16, salt + sample + str(io_len).encode("utf-8"))


class EncryptIO(IO):
    """Encrypt IO

    All methods are implied for stream cipher.
    Block cipher needs to imply by itself.
    """

    def __init__(self, io: IO, encrypt_password: bytes, total_origin_len: int):
        self._io = io

        # The length of unencrypted content of `self._io`
        self._io_len = total_len(self._io)

        # The offset of encrypted content of `self._io`
        self._offset = 0

        # The total length of all original (decrypted) content
        #
        # Only define it at __init__
        self._total_origin_len = total_origin_len

        self._encrypt_password = encrypt_password

        self._salt_for_head = generate_salt()
        self._encrypt_key_for_head, self._iv_for_head = generate_key_iv(
            encrypt_password, self._salt_for_head, 32, 16
        )

        # `self._encrypt_key`,`self._nonce_or_iv` is for cryptography
        self._salt = generate_salt()
        self._encrypt_key, self._nonce_or_iv = generate_key_iv(
            encrypt_password, self._salt, 32, 16
        )

        # Cryptography
        #
        # Instantiated at each subclass, here is for mypy
        self._crypto: Cryptography

        self._total_head: bytes
        self._total_head_len = len(self.total_head)

    def magic_code(self) -> bytes:
        raise NotImplementedError()

    def block_size(self) -> int:
        raise NotImplementedError()

    def reset(self):
        """Reset io and crypto"""
        self._io.seek(0, 0)
        self._offset = 0
        self._crypto.reset()

    @property
    def total_head(self) -> bytes:
        """
        Version 1:
        aes256cbc_encrypt(`BAIDUPCS_PY_CRYPTO_MAGIC_CODE 32bytes`, encrypt_password, random_bytes(16, encrypt_password)) +
        `encrypt algorithm code 1bytes` +
        `nonce or iv 16bytes` +
        `total_origin_len 8bytes`

        Version 2:
        aes256cbc_encrypt(
            `BAIDUPCS_PY_CRYPTO_MAGIC_CODE 32bytes` +
            `encrypt algorithm code 1bytes` +
            `nonce_or_iv 16bytes`
            `total_origin_len 8bytes`
        ) +
        `salt 8bytes`

        Version 3:
        aes256cbc_encrypt(
            `BAIDUPCS_PY_CRYPTO_MAGIC_CODE 32bytes` +
            `encrypt algorithm code 1bytes` +
            `salt 8bytes` + `random padding 8bytes`
            `total_origin_len 8bytes`
        ) +
        `salt 8bytes`
        """

        if hasattr(self, "_total_head"):
            return self._total_head

        assert len(self._salt) == 8

        ori_head = padding_key(
            BAIDUPCS_PY_CRYPTO_MAGIC_CODE
            + self.magic_code()
            + self._salt
            + random_sys_bytes(8)
            + u64_to_u8x8(self._total_origin_len),
            PADDED_ENCRYPT_HEAD_LEN,
            value=b"",
        )

        # AES256CBC Encrypt head
        self._total_head = (
            aes256cbc_encrypt(ori_head, self._encrypt_key_for_head, self._iv_for_head)
            + self._salt_for_head
        )
        return self._total_head

    def __len__(self) -> int:
        return self._total_head_len + self._io_len

    def read(self, size: int = -1) -> Optional[bytes]:
        """The read only imply to stream cipher"""

        if self._offset < self._total_head_len:
            if size < 0:
                data = self.total_head[self._offset :]
                io_buf = self._io.read()
                data += self._crypto.encrypt(io_buf)
            else:
                data = self.total_head[self._offset : self._offset + size]
                size -= len(data)
                if size > 0:
                    io_buf = self._io.read(size)
                    data += self._crypto.encrypt(io_buf)
        else:
            io_buf = self._io.read(size)
            data = self._crypto.encrypt(io_buf)
        self._offset += len(data)
        return data

    def seek(self, offset: int, whence: int = 0) -> int:
        if (
            not self.seekable()
            and offset == 0
            and whence == 0
            and self._total_head_len > 0
        ):
            self._crypto.reset()
            self._io.seek(0, 0)
            self._offset = 0
            return self._offset

        if not self.seekable():
            raise ValueError(
                f"{self._crypto.__class__.__name__} is not support seeking"
            )

        if whence == 0:
            if offset < 0:
                raise ValueError(f"Negative seek position {offset}")
            pass
        elif whence == 1:
            offset += self._offset
        elif whence == 2:
            offset = len(self) - offset

        io_offset = offset - self._total_head_len
        if io_offset > 0:
            io_len = self._io.seek(0, 2)
            io_offset = self._io.seek(min(io_offset, io_len), 0)
            self._offset = self._total_head_len + io_offset
        else:
            self._io.seek(0, 0)
            self._offset = max(0, offset)
        return self._offset

    def tell(self) -> int:
        return self._offset

    def seekable(self) -> bool:
        """For stream cipher"""

        return True

    def writable(self) -> bool:
        return False

    def close(self):
        self._io.close()


class SimpleEncryptIO(EncryptIO):
    MAGIC_CODE = b"\x00"

    BLOCK_SIZE = 16  # 16 bytes

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._crypto = SimpleCryptography(self._encrypt_key + self._nonce_or_iv)

    def magic_code(self) -> bytes:
        return self.MAGIC_CODE

    def block_size(self) -> int:
        return self.BLOCK_SIZE


class ChaCha20EncryptIO(EncryptIO):
    MAGIC_CODE = b"\x01"

    BLOCK_SIZE = 16  # 16 bytes

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._crypto = ChaCha20Cryptography(self._encrypt_key, self._nonce_or_iv)

    def reset(self):
        super().reset()

    def seekable(self) -> bool:
        return False

    def magic_code(self) -> bytes:
        return self.MAGIC_CODE

    def block_size(self) -> int:
        return self.BLOCK_SIZE


class AES256CBCEncryptIO(EncryptIO):
    MAGIC_CODE = b"\x02"

    BLOCK_SIZE = 16  # 16 bytes

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._crypto = AES256CBCCryptography(self._encrypt_key, self._nonce_or_iv)

        # The offset of encrypted origin content
        self._origin_io_offset = 0
        # Origin content cache
        self._origin_cache = bytearray()

        # Total length of final encrypted content
        self._encrypted_io_len = padding_size(self._io_len, self.BLOCK_SIZE)
        # Encrypted content cache
        self._encrypted_cache = bytearray()

        self._need_data_padded = self._total_origin_len != self._encrypted_io_len

    def magic_code(self) -> bytes:
        return self.MAGIC_CODE

    def block_size(self) -> int:
        return self.BLOCK_SIZE

    def reset(self):
        super().reset()
        self._origin_io_offset = 0
        self._origin_cache = bytearray()
        self._encrypted_cache = bytearray()

    def __len__(self):
        """Encrypted content length"""

        return self._total_head_len + self._encrypted_io_len

    def _read_block(self, size: int = -1):
        """Read encrypted block to cache"""

        # `self._encrypted_cache` has remains.
        if size > 0 and len(self._encrypted_cache) > size:
            return

        # `size` must be large or equal to the `size` as `size` be the times of `self.BLOCK_SIZE`
        if size > 0:
            size = padding_size(size, self.BLOCK_SIZE)

        data = self._io.read(size) or b""
        if not data:
            return

        self._origin_io_offset += len(data)
        self._origin_cache.extend(data)

        avail_ori_len = padding_size(
            len(self._origin_cache), self.BLOCK_SIZE, ceil=False
        )
        if avail_ori_len > 0:
            ori_cn = self._origin_cache[:avail_ori_len]
            self._origin_cache = self._origin_cache[avail_ori_len:]
        else:
            ori_cn = b""

        # The end encryption
        if self._origin_io_offset == self._total_origin_len:
            # Take all remainder
            if self._origin_cache:
                ori_cn += bytes(self._origin_cache)
                self._origin_cache.clear()

            # Padding
            if self._need_data_padded:
                ori_cn = pkcs7_padding(ori_cn, self.BLOCK_SIZE)

            enc_cn = self._crypto.encrypt(ori_cn)
            self._crypto.finalize()
        else:
            enc_cn = self._crypto.encrypt(ori_cn)

        self._encrypted_cache.extend(enc_cn)

    def _takeout(self, size: int = -1) -> bytes:
        """Take out from encrypted cache"""

        self._read_block(size)

        if size < 0:
            data = bytes(self._encrypted_cache)
            self._encrypted_cache.clear()
        else:
            data = bytes(self._encrypted_cache[:size])
            self._encrypted_cache = self._encrypted_cache[size:]
        return data

    def read(self, size: int = -1) -> Optional[bytes]:
        if self._offset < self._total_head_len:
            if size < 0:
                data = self.total_head[self._offset :]
                data += self._takeout()
            else:
                data = self.total_head[self._offset : self._offset + size]
                size -= len(data)
                if size > 0:
                    data += self._takeout(size)
        else:
            data = self._takeout(size)
        self._offset += len(data)
        return data

    def seekable(self) -> bool:
        return False


class DecryptIO(IO):
    def __init__(
        self, io: IO, encrypt_key: bytes, nonce_or_iv: bytes, total_origin_len: int
    ):
        self._io = io

        # The offset after ENCRYPT_HEAD_LEN
        # Must be equal ENCRYPT_HEAD_LEN
        #
        # When `self.set_io` be called, the `_io_init_offset` must be set to `io.tell()`
        self._io_init_offset = io.tell()

        # The length of decrypted content of `self._io`
        #
        # When `self.set_io` be called, the `_io_len` must be set to new io's len
        self._io_len = total_len(self._io)

        # The offset of decrypted content of `self._io`
        #
        # When `self.set_io` be called, the `_offset` must be set to `io.tell()`
        self._offset = 0

        self._encrypt_key = encrypt_key
        self._nonce_or_iv = nonce_or_iv

        # The total length of all original (decrypted) content
        #
        # Only define it at __init__
        self._total_origin_len = total_origin_len

        self._total_head_len = ENCRYPT_HEAD_LEN

    def __len__(self) -> int:
        return self._io_len - self._io_init_offset

    def set_io(self, io: IO):
        self._io = io
        self._io_init_offset = io.tell()
        self._offset = 0
        self._io_len = total_len(self._io)

    def read(self, size: int = -1) -> Optional[bytes]:
        data = self._io.read(size)
        if not data:
            return b""
        self._offset += len(data)
        return data

    def seek(self, offset: int, whence: int = 0) -> int:
        # ChaCha20 and AES are depended on previous decrypted
        if not self.seekable():
            raise ValueError(f"{self.__class__.__name__} is not support seeking")

        if whence == 0:
            if offset < 0:
                raise ValueError(f"Negative seek position {offset}")
            pass
        elif whence == 1:
            offset += self._offset
        elif whence == 2:
            offset = len(self) - offset

        offset = max(0, offset)
        offset = min(len(self), offset)
        self._io.seek(offset + self._io_init_offset, 0)
        self._offset = offset
        return offset

    def tell(self) -> int:
        return self._offset

    def seekable(self) -> bool:
        return True

    def writable(self) -> bool:
        return False

    def close(self):
        self._io.close()


class SimpleDecryptIO(DecryptIO):
    def __init__(self, io: IO, encrypt_key: bytes, nonce: bytes, total_origin_len: int):
        super().__init__(io, encrypt_key, nonce, total_origin_len)

        self._crypto = SimpleCryptography(self._encrypt_key + self._nonce_or_iv)

    def read(self, size: int = -1) -> Optional[bytes]:
        data = self._io.read(size)
        if not data:
            return b""
        self._offset += len(data)
        return self._crypto.decrypt(data)

    def seekable(self) -> bool:
        return True


class ChaCha20DecryptIO(DecryptIO):
    BLOCK_SIZE = 16  # 16 bytes

    def __init__(self, io: IO, encrypt_key: bytes, nonce: bytes, total_origin_len: int):
        encrypt_key = padding_key(encrypt_key, self.BLOCK_SIZE * 2)
        nonce = padding_key(nonce, self.BLOCK_SIZE)
        super().__init__(io, encrypt_key, nonce, total_origin_len)

        self._crypto = ChaCha20Cryptography(self._encrypt_key, self._nonce_or_iv)

    def read(self, size: int = -1) -> Optional[bytes]:
        data = self._io.read(size)
        if not data:
            return b""
        self._offset += len(data)
        return self._crypto.decrypt(data)

    def seekable(self) -> bool:
        return False


class AES256CBCDecryptIO(DecryptIO):
    BLOCK_SIZE = 16  # 16 bytes

    def __init__(self, io: IO, encrypt_key: bytes, iv: bytes, total_origin_len: int):
        encrypt_key = padding_key(encrypt_key, self.BLOCK_SIZE * 2)
        iv = padding_key(iv, self.BLOCK_SIZE)
        super().__init__(io, encrypt_key, iv, total_origin_len)

        self._crypto = AES256CBCCryptography(self._encrypt_key, self._nonce_or_iv)

        self._encrypted_io_offset = 0
        self._encrypted_io_len = padding_size(total_origin_len, self.BLOCK_SIZE)
        self._encrypted_data_padded = total_origin_len != self._encrypted_io_len

        self._encrypted_cache = bytearray()
        self._decrypted_cache = bytearray()

    def __len__(self) -> int:
        """The decrypted content length of `self._io`"""

        avail_enc_len = padding_size(
            self._io_len - self._io_init_offset, self.BLOCK_SIZE, ceil=False
        )
        return avail_enc_len

    def _read_block(self, size: int = -1):
        """Read encrypted block to cache"""

        # `self._decrypted_cache` has remains.
        if size > 0 and len(self._decrypted_cache) > size:
            return

        # `size` must be large or equal to the `size` as `size` be the times of `self.BLOCK_SIZE`
        if size > 0:
            size = padding_size(size, self.BLOCK_SIZE)

        data = self._io.read(size)
        if not data:
            return

        self._encrypted_io_offset += len(data)
        self._encrypted_cache.extend(data)

        avail_enc_len = padding_size(
            len(self._encrypted_cache), self.BLOCK_SIZE, ceil=False
        )
        if avail_enc_len > 0:
            enc_cn = bytes(self._encrypted_cache[:avail_enc_len])
            self._encrypted_cache = self._encrypted_cache[avail_enc_len:]
        else:
            enc_cn = b""

        # The end decryption
        if self._encrypted_io_offset == self._encrypted_io_len:
            if self._encrypted_cache:
                enc_cn += bytes(self._encrypted_cache)
                self._encrypted_cache.clear()

            dec_cn = self._crypto.decrypt(enc_cn)

            if self._encrypted_data_padded:
                dec_cn = pkcs7_unpadding(dec_cn, self.BLOCK_SIZE)

            self._crypto.finalize()
        else:
            dec_cn = self._crypto.decrypt(enc_cn)

        self._decrypted_cache.extend(dec_cn)

    def read(self, size: int = -1) -> Optional[bytes]:
        self._read_block(size)

        if size < 0:
            data = bytes(self._decrypted_cache)
            self._decrypted_cache.clear()
        else:
            data = bytes(self._decrypted_cache[:size])
            self._decrypted_cache = self._decrypted_cache[size:]
        self._offset += len(data)
        return data

    def seekable(self) -> bool:
        return False


def parse_head(head: bytes) -> Tuple[bytes, bytes, bytes, bytes]:
    i = len(BAIDUPCS_PY_CRYPTO_MAGIC_CODE)
    return (
        head[:i],
        head[i : i + 1],  # magic_code
        head[i + 1 : i + 1 + 16],  # salt + random padding
        head[i + 1 + 16 : i + 1 + 16 + 8],  # total_origin_len
    )


def _decryptio_version1(
    total_head: bytes, io: IO, encrypt_password: bytes
) -> Optional[DecryptIO]:
    encrypt_password = padding_key(encrypt_password, 32)

    if len(total_head) < ENCRYPT_HEAD_LEN:
        return None

    # Version 1
    b_mc, magic_code, nonce_or_iv, total_origin_len_ = parse_head(total_head)
    b_mc = aes256cbc_decrypt(b_mc, encrypt_password, random_bytes(16, encrypt_password))
    total_origin_len = u8x8_to_u64(total_origin_len_)

    if b_mc != BAIDUPCS_PY_CRYPTO_MAGIC_CODE:
        return None

    if magic_code == SimpleEncryptIO.MAGIC_CODE:
        return SimpleDecryptIO(io, encrypt_password, b"", total_origin_len)
    elif magic_code == ChaCha20EncryptIO.MAGIC_CODE:
        return ChaCha20DecryptIO(io, encrypt_password, nonce_or_iv, total_origin_len)
    elif magic_code == AES256CBCEncryptIO.MAGIC_CODE:
        return AES256CBCDecryptIO(io, encrypt_password, nonce_or_iv, total_origin_len)
    else:
        logging.warning(f"Unknown magic_code: {magic_code!r}")
        return None


def _decryptio_version3(
    total_head: bytes, io: IO, encrypt_password: bytes
) -> Optional[DecryptIO]:
    if len(total_head) < PADDED_ENCRYPT_HEAD_WITH_SALT_LEN:
        return None

    enc_head, salt_for_head = (
        total_head[:PADDED_ENCRYPT_HEAD_LEN],
        total_head[PADDED_ENCRYPT_HEAD_LEN:],
    )

    encrypt_key_for_head, iv_for_head = generate_key_iv(
        encrypt_password, salt_for_head, 32, 16
    )

    head = aes256cbc_decrypt(enc_head, encrypt_key_for_head, iv_for_head)

    b_mc, magic_code, padding_salt, total_origin_len_ = parse_head(head)
    total_origin_len = u8x8_to_u64(total_origin_len_)

    salt = padding_salt[:8]
    encrypt_key, nonce_or_iv = generate_key_iv(encrypt_password, salt, 32, 16)

    if b_mc != BAIDUPCS_PY_CRYPTO_MAGIC_CODE:
        return None

    eio: DecryptIO
    if magic_code == SimpleEncryptIO.MAGIC_CODE:
        eio = SimpleDecryptIO(io, encrypt_key, nonce_or_iv, total_origin_len)
    elif magic_code == ChaCha20EncryptIO.MAGIC_CODE:
        eio = ChaCha20DecryptIO(io, encrypt_key, nonce_or_iv, total_origin_len)
    elif magic_code == AES256CBCEncryptIO.MAGIC_CODE:
        eio = AES256CBCDecryptIO(io, encrypt_key, nonce_or_iv, total_origin_len)
    else:
        logging.warning(f"Unknown magic_code: {magic_code!r}")
        return None

    eio._total_head_len = PADDED_ENCRYPT_HEAD_WITH_SALT_LEN
    return eio


def to_decryptio(io: IO, encrypt_password: bytes):
    if not encrypt_password:
        return io

    total_head = io.read(ENCRYPT_HEAD_LEN)
    eio = _decryptio_version1(total_head, io, encrypt_password)
    if eio is not None:
        return eio

    # No support Version 2

    total_head += io.read(PADDED_ENCRYPT_HEAD_WITH_SALT_LEN - ENCRYPT_HEAD_LEN)
    eio = _decryptio_version3(total_head, io, encrypt_password)
    if eio is not None:
        return eio

    io.seek(0, 0)
    return io


class AutoDecryptRequest:
    def __init__(
        self,
        method: str,
        url: str,
        headers: Optional[Dict[str, str]] = None,
        max_chunk_size: int = MAX_CHUNK_SIZE,
        encrypt_password: bytes = b"",
        **kwargs,
    ):
        kwargs["stream"] = True

        self._method = method
        self._url = url
        self._headers = headers
        self._session = requests.session()
        self._kwargs = kwargs
        self._max_chunk_size = max_chunk_size
        self._encrypt_password = encrypt_password
        self._dio: Optional[IO] = None
        self._has_encrypted = False
        self._total_head_len = 0
        self._decrypted_count = 0
        self._parsed = False

        # Rapid Upload Infos
        #
        # Total raw content length
        self._content_length: Optional[int] = None
        # Total raw content md5
        self._content_md5: Optional[str] = None
        # Total raw content crc32
        self._content_crc32: Optional[int] = None

    def _init(self):
        """Initiate request info"""

        if not self._parsed:
            self._parse_crypto()

        if self._content_length is None:
            # To invoke `self._parse_rapid_upload_info`
            with self._request((0, 1)):
                pass

    @property
    def content_length(self) -> int:
        """Remote content length"""

        self._init()
        assert self._content_length
        return self._content_length

    @property
    def content_md5(self) -> Optional[str]:
        """Remote content length"""

        self._init()
        assert self._content_length
        return self._content_md5

    @property
    def content_crc32(self) -> Optional[int]:
        """Remote content length"""

        self._init()
        assert self._content_length
        return self._content_crc32

    def _parse_rapid_upload_info(self, resp: Response):
        if not resp.ok or self._content_length:
            return

        headers = resp.headers
        if headers.get("x-bs-file-size"):
            self._content_length = int(headers["x-bs-file-size"])
        if not self._content_length and headers.get("Content-Range"):
            self._content_length = int(headers["Content-Range"].split("/")[-1])

        if headers.get("Content-MD5"):
            self._content_md5 = headers["Content-MD5"]

        if headers.get("x-bs-meta-crc32"):
            self._content_crc32 = int(headers["x-bs-meta-crc32"])

    def _parse_crypto(self):
        if self._encrypt_password:
            with self._request((0, PADDED_ENCRYPT_HEAD_WITH_SALT_LEN - 1)) as resp:
                raw_data = resp.raw.read()
                if len(raw_data) == PADDED_ENCRYPT_HEAD_WITH_SALT_LEN:
                    self._dio = to_decryptio(BytesIO(raw_data), self._encrypt_password)
                    self._has_encrypted = isinstance(self._dio, DecryptIO)
                    self._total_head_len = (
                        cast(DecryptIO, self._dio)._total_head_len
                        if self._has_encrypted
                        else 0
                    )
        self._parsed = True

    def _request(self, _range: Tuple[int, int]) -> Response:
        headers = dict(self._headers or {})
        headers["Range"] = "bytes={}-{}".format(*_range)

        while True:
            try:
                resp = self._session.request(
                    self._method, self._url, headers=headers, **self._kwargs
                )
                if not resp.ok:
                    logger.warning(
                        "`%s._request` request error: status_code: %s, body: %s",
                        self.__class__.__name__,
                        resp.status_code,
                        resp.content[:1000],
                    )

                    # Error: 31626: user is not authorized, hitcode:122
                    # Error: 31326: user is not authorized, hitcode:117
                    # Request is temporally blocked.
                    # This error occurs when url is from `BaiduPCS.download_link(..., pcs=True)`
                    if b"user is not authorized" in resp.content:
                        time.sleep(2)
                        continue
                self._parse_rapid_upload_info(resp)
                break
            except Exception as err:
                logger.warning(
                    "`%s._request` request error: %s",
                    self.__class__.__name__,
                    err,
                )
                raise IOError(f"{self.__class__.__name__} - Request Error") from err

        if not resp.ok:
            raise IOError(
                f"{self.__class__.__name__} - Response is not ok: "
                f"status_code: {resp.status_code}, body: {resp.content[:1000]}"
            )

        return resp

    def rangeable(self) -> bool:
        """Is support uncontinue range request?"""

        self._init()

        if self._dio is None:
            return True
        return self._dio.seekable()

    def __len__(self) -> int:
        """The decrypted data length"""

        self._init()

        assert self._content_length
        return self._content_length - self._total_head_len

    def read(self, _range: Tuple[int, int]) -> Generator[bytes, None, None]:
        self._init()

        start, end = _range
        if end < 0:
            end = len(self)
        end = min(end, len(self))

        if end - start == 0:
            return None

        if start != self._decrypted_count and not self.rangeable():
            raise IndexError(
                "ChaCha20 must decoded data with continue data."
                f"start: {start}, decrypted count: {self._decrypted_count}"
            )

        ranges = self._split_chunk(
            start + self._total_head_len, end + self._total_head_len
        )
        for _rg in ranges:
            with self._request(_rg) as resp:
                stream = resp.raw
                while True:
                    buf = stream.read(READ_SIZE)
                    if not buf:
                        break
                    self._decrypted_count += len(buf)
                    bio = BytesIO(buf)
                    if self._dio is not None and isinstance(self._dio, DecryptIO):
                        self._dio.set_io(bio)
                        yield self._dio.read() or b""
                    else:
                        yield buf

    def _split_chunk(self, start: int, end: int) -> List[Tuple[int, int]]:
        """Split the chunks for range header

        Echo chunk has the length at most `self._max_chunk_size`.
        """

        left = list(range(start, end, self._max_chunk_size))
        left.append(end)
        right = [i - 1 for i in left]
        return [(s, e) for s, e in zip(left[:-1], right[1:])]


class RangeRequestIO(IO):
    def __init__(
        self,
        method: str,
        url: str,
        headers: Optional[Dict[str, str]] = None,
        max_chunk_size: int = MAX_CHUNK_SIZE,
        callback: Callable[..., None] = None,
        encrypt_password: bytes = b"",
        **kwargs,
    ):
        kwargs["stream"] = True

        self._method = method
        self._url = url
        self._headers = headers
        self._kwargs = kwargs
        self._max_chunk_size = max_chunk_size
        self._callback = callback
        self._encrypt_password = encrypt_password

        self.reset()

    def reset(self):
        self._offset = 0
        self._auto_decrypt_request = AutoDecryptRequest(
            self._method,
            self._url,
            headers=self._headers,
            max_chunk_size=self._max_chunk_size,
            encrypt_password=self._encrypt_password,
            **self._kwargs,
        )

    def __len__(self) -> int:
        return len(self._auto_decrypt_request)

    def read(self, size: int = -1) -> Optional[bytes]:
        if size == 0:
            return b""

        if size == -1:
            size = len(self) - self._offset

        start, end = self._offset, self._offset + size

        buf = b""
        for b in self._auto_decrypt_request.read((start, end)):
            buf += b
            self._offset += len(b)
            # Call callback
            if self._callback:
                self._callback(self._offset)
        return buf

    def seek(self, offset: int, whence: int = 0) -> int:
        if whence == 0:
            self._offset = offset
        elif whence == 1:
            self._offset += offset
        elif whence == 2:
            self._offset = len(self) - offset

        self._offset = max(0, self._offset)
        self._offset = min(len(self), self._offset)
        return self._offset

    def tell(self) -> int:
        return self._offset

    def seekable(self) -> bool:
        return self._auto_decrypt_request.rangeable()

    def writable(self) -> bool:
        return False

    def close(self):
        pass


class EncryptType(Enum):
    No = "No"
    Simple = "Simple"
    ChaCha20 = "ChaCha20"
    AES256CBC = "AES256CBC"

    def encrypt_io(self, io: IO, encrypt_password: bytes) -> IO:
        io_len = total_len(io)
        if self == EncryptType.No:
            return io
        elif self == EncryptType.Simple:
            return SimpleEncryptIO(io, encrypt_password, io_len)
        elif self == EncryptType.ChaCha20:
            return ChaCha20EncryptIO(io, encrypt_password, io_len)
        elif self == EncryptType.AES256CBC:
            return AES256CBCEncryptIO(io, encrypt_password, io_len)
        else:
            raise ValueError(f"Unknown EncryptType: {self}")


def reset_encrypt_io(io: Union[IO, EncryptIO]):
    if isinstance(io, EncryptIO):
        io.reset()
    else:
        io.seek(0, 0)



import functools
import os
import time
from concurrent.futures import ThreadPoolExecutor, as_completed
from enum import Enum
from io import BytesIO
from pathlib import Path
from threading import Semaphore
from typing import IO, List, Optional, Tuple

from requests_toolbelt import MultipartEncoderMonitor
from rich import print
from rich.box import SIMPLE
from rich.progress import TaskID
from rich.table import Table
from rich.text import Text

from ..common import constant
from ..common.concurrent import retry, sure_release
from ..common.constant import CPU_NUM
from ..common.event import KeyboardMonitor, KeyHandler
from ..common.io import (
    EncryptType,
    rapid_upload_params,
    reset_encrypt_io,
    total_len,
)
from ..common.localstorage import save_rapid_upload_info
from ..common.path import exists, is_file, walk
from ..common.progress_bar import _progress, progress_task_exists
from ..pcs import BaiduPCSApi, FromTo
from ..pcs.errors import BaiduPCSError
from .log import get_logger

logger = get_logger(__name__)

# If slice size >= 100M, the rate of uploading will be much lower.
DEFAULT_SLICE_SIZE = 30 * constant.OneM


UPLOAD_STOP = False

_rapiduploadinfo_file: Optional[str] = None


def _wait_start():
    while True:
        if UPLOAD_STOP:
            time.sleep(1)
        else:
            break


def _toggle_stop(*args, **kwargs):
    global UPLOAD_STOP
    UPLOAD_STOP = not UPLOAD_STOP
    if UPLOAD_STOP:
        print("[i yellow]Uploading stop[/i yellow]")
    else:
        print("[i yellow]Uploading continue[/i yellow]")


# Pass "p" to toggle uploading start/stop
KeyboardMonitor.register(KeyHandler("p", callback=_toggle_stop))


def to_remotepath(sub_path: str, remotedir: str) -> str:
    return (Path(remotedir) / sub_path).as_posix()


def from_tos(localpaths: List[str], remotedir: str) -> List[FromTo]:
    """Find all localpaths and their corresponded remotepath"""

    ft: List[FromTo] = []
    for localpath in localpaths:
        if not exists(Path(localpath)):
            continue

        if is_file(Path(localpath)):
            remotepath = to_remotepath(os.path.basename(localpath), remotedir)
            ft.append(FromTo(localpath, remotepath))
        else:
            parents_num = max(len(Path(localpath).parts) - 1, 0)
            for sub_path in walk(Path(localpath)):
                relative_path = Path(*Path(sub_path).parts[parents_num:]).as_posix()
                remotepath = to_remotepath(relative_path, remotedir)
                ft.append(FromTo(sub_path, remotepath))
    return ft


class UploadType(Enum):
    """Upload Type

    One: Upload the slices of one file concurrently
    Many: Upload files concurrently
    """

    One = 1
    Many = 2


def _handle_deadly_error(err, fail_count):
    # If following errors occur, we need to re-upload
    if isinstance(err, BaiduPCSError) and (
        err.error_code == 31352  # commit superfile2 failed
        or err.error_code == 31363  # block miss in superfile2
        or err.error_code == 31062  # 文件名非法
        or err.error_code == 31112  # 超出配额
    ):
        logger.warning(
            "Deadly error: %s, fail_count: %s",
            err,
            fail_count,
            exc_info=err,
        )
        raise err


# remotedir must be a directory
def upload(
    api: BaiduPCSApi,
    from_to_list: List[FromTo],
    upload_type: UploadType = UploadType.One,
    ondup: str = "overwrite",
    encrypt_password: bytes = b"",
    encrypt_type: EncryptType = EncryptType.No,
    max_workers: int = CPU_NUM,
    slice_size: int = DEFAULT_SLICE_SIZE,
    ignore_existing: bool = True,
    show_progress: bool = True,
    rapiduploadinfo_file: Optional[str] = None,
    user_id: Optional[int] = None,
    user_name: Optional[str] = None,
    check_md5: bool = False,
):
    """Upload from_tos

    Args:
        upload_type (UploadType): the way of uploading.
        max_workers (int): The number of concurrent workers.
        slice_size (int): The size of slice for uploading slices.
        ignore_existing (bool): Ignoring these localpath which of remotepath exist.
        show_progress (bool): Show uploading progress.

        check_md5 (bool): To fix the content md5 after `combine_slices`
            `combine_slices` always does not return correct content md5. To fix it,
            we need to use `rapid_upload_file` re-upload the content.
            Warning, if content length is large, it could take some minutes,
            e.g. it takes 5 minutes about 2GB.
    """

    logger.debug(
        "======== Uploading start ========\n-> UploadType: %s\n-> Size of from_to_list: %s",
        upload_type,
        len(from_to_list),
    )

    global _rapiduploadinfo_file
    if _rapiduploadinfo_file is None:
        _rapiduploadinfo_file = rapiduploadinfo_file

    if upload_type == UploadType.One:
        upload_one_by_one(
            api,
            from_to_list,
            ondup,
            max_workers=max_workers,
            encrypt_password=encrypt_password,
            encrypt_type=encrypt_type,
            slice_size=slice_size,
            ignore_existing=ignore_existing,
            show_progress=show_progress,
            user_id=user_id,
            user_name=user_name,
            check_md5=check_md5,
        )
    elif upload_type == UploadType.Many:
        upload_many(
            api,
            from_to_list,
            ondup,
            max_workers=max_workers,
            encrypt_password=encrypt_password,
            encrypt_type=encrypt_type,
            slice_size=slice_size,
            ignore_existing=ignore_existing,
            show_progress=show_progress,
            user_id=user_id,
            user_name=user_name,
            check_md5=check_md5,
        )


def _init_encrypt_io(
    api: BaiduPCSApi,
    localpath: str,
    remotepath: str,
    encrypt_password: bytes = b"",
    encrypt_type: EncryptType = EncryptType.No,
    ignore_existing: bool = True,
    task_id: Optional[TaskID] = None,
) -> Optional[Tuple[IO, int, int, int]]:
    assert exists(Path(localpath)), f"`{localpath}` does not exist"

    if ignore_existing:
        try:
            if api.exists(remotepath):
                print(f"`{remotepath}` already exists.")
                logger.debug("`_init_encrypt_io`: remote file already exists")
                if task_id is not None and progress_task_exists(task_id):
                    _progress.remove_task(task_id)
                return None
        except Exception as err:
            if task_id is not None and progress_task_exists(task_id):
                _progress.remove_task(task_id)
            raise err

    stat = Path(localpath).stat()
    local_ctime, local_mtime = int(stat.st_ctime), int(stat.st_mtime)

    encrypt_io = encrypt_type.encrypt_io(open(localpath, "rb"), encrypt_password)
    # IO Length
    encrypt_io_len = total_len(encrypt_io)

    logger.debug(
        "`_init_encrypt_io`: encrypt_type: %s, localpath: %s, remotepath: %s, encrypt_io_len: %s",
        encrypt_type,
        localpath,
        remotepath,
        encrypt_io_len,
    )

    return (encrypt_io, encrypt_io_len, local_ctime, local_mtime)


def _rapid_upload(
    api: BaiduPCSApi,
    localpath: str,
    remotepath: str,
    slice256k_md5: str,
    content_md5: str,
    content_crc32: int,
    io_len: int,
    local_ctime: int,
    local_mtime: int,
    ondup: str,
    encrypt_password: bytes = b"",
    encrypt_type: EncryptType = EncryptType.No,
    task_id: Optional[TaskID] = None,
    user_id: Optional[int] = None,
    user_name: Optional[str] = None,
) -> bool:
    logger.debug("`_rapid_upload`: rapid_upload starts")
    try:
        api.rapid_upload_file(
            slice256k_md5,
            content_md5,
            0,  # not needed
            io_len,
            remotepath,
            local_ctime=local_ctime,
            local_mtime=local_mtime,
            ondup=ondup,
        )

        if _rapiduploadinfo_file:
            save_rapid_upload_info(
                _rapiduploadinfo_file,
                slice256k_md5,
                content_md5,
                content_crc32,
                io_len,
                localpath=localpath,
                remotepath=remotepath,
                encrypt_password=encrypt_password,
                encrypt_type=encrypt_type.value,
                user_id=user_id,
                user_name=user_name,
            )

        if task_id is not None and progress_task_exists(task_id):
            _progress.update(task_id, completed=io_len)
            _progress.remove_task(task_id)

        logger.debug("`_rapid_upload`: rapid_upload success, task_id: %s", task_id)
        return True
    except BaiduPCSError as err:
        logger.warning("`_rapid_upload`: rapid_upload fails")

        if (
            err.error_code != 31079
        ):  # 31079: '未找到文件MD5，请使用上传API上传整个文件。'
            if task_id is not None and progress_task_exists(task_id):
                _progress.remove_task(task_id)

            logger.warning("`_rapid_upload`: unknown error: %s", err)
            raise err
        else:
            logger.debug("`_rapid_upload`: %s, no exist in remote", localpath)

            if task_id is not None and progress_task_exists(task_id):
                _progress.reset(task_id)

        return False


@retry(20, except_callback=_handle_deadly_error)
def _combine_slices(
    api: BaiduPCSApi,
    remotepath: str,
    slice_md5s: List[str],
    local_ctime: int,
    local_mtime: int,
    ondup: str,
):
    api.combine_slices(
        slice_md5s,
        remotepath,
        local_ctime=local_ctime,
        local_mtime=local_mtime,
        ondup=ondup,
    )


def upload_one_by_one(
    api: BaiduPCSApi,
    from_to_list: List[FromTo],
    ondup: str,
    max_workers: int = CPU_NUM,
    encrypt_password: bytes = b"",
    encrypt_type: EncryptType = EncryptType.No,
    slice_size: int = DEFAULT_SLICE_SIZE,
    ignore_existing: bool = True,
    show_progress: bool = True,
    user_id: Optional[int] = None,
    user_name: Optional[str] = None,
    check_md5: bool = False,
):
    """Upload files one by one with uploading the slices concurrently"""

    for from_to in from_to_list:
        task_id = None
        if show_progress:
            task_id = _progress.add_task("upload", start=False, title=from_to.from_)
        upload_file_concurrently(
            api,
            from_to,
            ondup,
            max_workers=max_workers,
            encrypt_password=encrypt_password,
            encrypt_type=encrypt_type,
            slice_size=slice_size,
            ignore_existing=ignore_existing,
            task_id=task_id,
            user_id=user_id,
            user_name=user_name,
            check_md5=check_md5,
        )

    logger.debug("======== Uploading end ========")


@retry(
    -1,
    except_callback=lambda err, fail_count: (
        _handle_deadly_error(err, fail_count),
        logger.warning(
            "`upload_file_concurrently`: fails: error: %s, fail_count: %s",
            err,
            fail_count,
            exc_info=err,
        ),
    ),
)
def upload_file_concurrently(
    api: BaiduPCSApi,
    from_to: FromTo,
    ondup: str,
    max_workers: int = CPU_NUM,
    encrypt_password: bytes = b"",
    encrypt_type: EncryptType = EncryptType.No,
    slice_size: int = DEFAULT_SLICE_SIZE,
    ignore_existing: bool = True,
    task_id: Optional[TaskID] = None,
    user_id: Optional[int] = None,
    user_name: Optional[str] = None,
    check_md5: bool = False,
):
    """Uploading one file by uploading it's slices concurrently"""

    localpath, remotepath = from_to

    info = _init_encrypt_io(
        api,
        localpath,
        remotepath,
        encrypt_password=encrypt_password,
        encrypt_type=encrypt_type,
        ignore_existing=ignore_existing,
        task_id=task_id,
    )
    if not info:
        return
    encrypt_io, encrypt_io_len, local_ctime, local_mtime = info

    # Progress bar
    if task_id is not None and progress_task_exists(task_id):
        _progress.update(task_id, total=encrypt_io_len)
        _progress.start_task(task_id)

    slice_completed = 0
    slice_completeds = {}  # current i-th index slice completed size

    def callback_for_slice(idx: int, monitor: MultipartEncoderMonitor):
        if task_id is not None and progress_task_exists(task_id):
            slice_completeds[idx] = monitor.bytes_read
            current_compledted: int = sum(list(slice_completeds.values()))
            _progress.update(task_id, completed=slice_completed + current_compledted)

    slice256k_md5 = ""
    content_md5 = ""
    content_crc32 = 0

    if encrypt_type == EncryptType.No and encrypt_io_len > 256 * constant.OneK:
        # Rapid Upload
        slice256k_md5, content_md5, content_crc32, _ = rapid_upload_params(encrypt_io)
        ok = _rapid_upload(
            api,
            localpath,
            remotepath,
            slice256k_md5,
            content_md5,
            content_crc32,
            encrypt_io_len,
            local_ctime,
            local_mtime,
            ondup,
            encrypt_password=encrypt_password,
            encrypt_type=encrypt_type,
            task_id=task_id,
            user_id=user_id,
            user_name=user_name,
        )
        if ok:
            return

    try:
        # Upload file slice
        logger.debug("`upload_file_concurrently`: upload_slice starts")

        reset_encrypt_io(encrypt_io)

        completed_slice_md5s = []

        def upload_slice(item: Tuple[int, IO]):
            idx, io = item

            # Retry upload until success
            slice_md5 = retry(
                -1,
                except_callback=lambda err, fail_count: (
                    _handle_deadly_error(err, fail_count),
                    io.seek(0, 0),
                    logger.warning(
                        "`upload_file_concurrently`: error: %s, fail_count: %s",
                        err,
                        fail_count,
                        exc_info=err,
                    ),
                    _wait_start(),
                ),
            )(api.upload_slice)(io, callback=functools.partial(callback_for_slice, idx))

            slice_completeds.pop(idx)
            completed_slice_md5s.append((idx, slice_md5))

            nonlocal slice_completed
            slice_completed += total_len(io)

        semaphore = Semaphore(max_workers)
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futs = []
            offset = 0
            idx = 0
            while True:
                semaphore.acquire()

                size = min(slice_size, encrypt_io_len - offset)
                if idx != 0 and size == 0:
                    break

                data = encrypt_io.read(size)
                io = BytesIO(data or b"")

                fut = executor.submit(sure_release, semaphore, upload_slice, (idx, io))
                futs.append(fut)

                idx += 1
                offset += size

            as_completed(futs)

        completed_slice_md5s.sort()
        slice_md5s = [md5 for _, md5 in completed_slice_md5s]

        # Combine slices
        _combine_slices(
            api,
            remotepath,
            slice_md5s,
            local_ctime,
            local_mtime,
            ondup,
        )

        logger.debug(
            "`upload_file_concurrently`: upload_slice and combine_slices success, task_id: %s",
            task_id,
        )

        # `combine_slices` can not get right content md5.
        # We need to check whether server updates by hand.
        if check_md5:
            _check_md5(
                api,
                localpath,
                remotepath,
                slice256k_md5,
                content_md5,
                content_crc32,
                encrypt_io_len,
                encrypt_password=encrypt_password,
                encrypt_type=encrypt_type.value,
                user_id=user_id,
                user_name=user_name,
            )

        if task_id is not None and progress_task_exists(task_id):
            _progress.remove_task(task_id)
    except Exception as err:
        logger.warning("`upload_file_concurrently`: error: %s", err)
        raise err
    finally:
        encrypt_io.close()
        if task_id is not None and progress_task_exists(task_id):
            _progress.reset(task_id)


def upload_many(
    api: BaiduPCSApi,
    from_to_list: List[FromTo],
    ondup: str = "overwrite",
    encrypt_password: bytes = b"",
    encrypt_type: EncryptType = EncryptType.No,
    max_workers: int = CPU_NUM,
    slice_size: int = DEFAULT_SLICE_SIZE,
    ignore_existing: bool = True,
    show_progress: bool = True,
    rapiduploadinfo_file: Optional[str] = None,
    user_id: Optional[int] = None,
    user_name: Optional[str] = None,
    check_md5: bool = False,
):
    """Upload files concurrently that one file is with one connection"""

    excepts = {}
    semaphore = Semaphore(max_workers)
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        futs = {}
        for idx, from_to in enumerate(from_to_list):
            semaphore.acquire()
            task_id = None
            if show_progress:
                task_id = _progress.add_task("upload", start=False, title=from_to.from_)

            logger.debug("`upload_many`: Upload: index: %s, task_id: %s", idx, task_id)

            fut = executor.submit(
                sure_release,
                semaphore,
                upload_file,
                api,
                from_to,
                ondup,
                encrypt_password=encrypt_password,
                encrypt_type=encrypt_type,
                slice_size=slice_size,
                ignore_existing=ignore_existing,
                task_id=task_id,
                user_id=user_id,
                user_name=user_name,
                check_md5=check_md5,
            )
            futs[fut] = from_to

        for fut in as_completed(futs):
            e = fut.exception()
            if e is not None:
                from_to = futs[fut]
                excepts[from_to] = e

    logger.debug("======== Uploading end ========")

    # Summary
    if excepts:
        table = Table(title="Upload Error", box=SIMPLE, show_edge=False)
        table.add_column("From", justify="left", overflow="fold")
        table.add_column("To", justify="left", overflow="fold")
        table.add_column("Error", justify="left")

        for from_to, e in sorted(excepts.items()):
            table.add_row(from_to.from_, Text(str(e), style="red"))

        _progress.console.print(table)


@retry(
    -1,
    except_callback=lambda err, fail_count: (
        _handle_deadly_error(err, fail_count),
        logger.warning(
            "`upload_file`: fails: error: %s, fail_count: %s",
            err,
            fail_count,
            exc_info=err,
        ),
    ),
)
def upload_file(
    api: BaiduPCSApi,
    from_to: FromTo,
    ondup: str,
    encrypt_password: bytes = b"",
    encrypt_type: EncryptType = EncryptType.No,
    slice_size: int = DEFAULT_SLICE_SIZE,
    ignore_existing: bool = True,
    task_id: Optional[TaskID] = None,
    user_id: Optional[int] = None,
    user_name: Optional[str] = None,
    check_md5: bool = False,
):
    """Upload one file with one connection"""

    _wait_start()

    localpath, remotepath = from_to

    info = _init_encrypt_io(
        api,
        localpath,
        remotepath,
        encrypt_password=encrypt_password,
        encrypt_type=encrypt_type,
        ignore_existing=ignore_existing,
        task_id=task_id,
    )
    if not info:
        return
    encrypt_io, encrypt_io_len, local_ctime, local_mtime = info

    # Progress bar
    if task_id is not None and progress_task_exists(task_id):
        _progress.update(task_id, total=encrypt_io_len)
        _progress.start_task(task_id)

    slice_completed = 0

    def callback_for_slice(monitor: MultipartEncoderMonitor):
        if task_id is not None and progress_task_exists(task_id):
            _progress.update(task_id, completed=slice_completed + monitor.bytes_read)

    slice256k_md5 = ""
    content_md5 = ""
    content_crc32 = 0

    if encrypt_type == EncryptType.No and encrypt_io_len > 256 * constant.OneK:
        # Rapid Upload
        slice256k_md5, content_md5, content_crc32, _ = rapid_upload_params(encrypt_io)
        ok = _rapid_upload(
            api,
            localpath,
            remotepath,
            slice256k_md5,
            content_md5,
            content_crc32,
            encrypt_io_len,
            local_ctime,
            local_mtime,
            ondup,
            encrypt_password=encrypt_password,
            encrypt_type=encrypt_type,
            task_id=task_id,
            user_id=user_id,
            user_name=user_name,
        )
        if ok:
            return

    try:
        # Upload file slice
        logger.debug("`upload_file`: upload_slice starts")

        slice_md5s = []
        reset_encrypt_io(encrypt_io)

        idx = 0
        while True:
            _wait_start()

            logger.debug(
                "`upload_file`: upload_slice: slice_completed: %s", slice_completed
            )

            size = min(slice_size, encrypt_io_len - slice_completed)
            if idx != 0 and size == 0:
                break

            data = encrypt_io.read(size) or b""
            io = BytesIO(data)

            logger.debug(
                "`upload_file`: upload_slice: size should be %s == %s", size, len(data)
            )

            # Retry upload until success
            slice_md5 = retry(
                -1,
                except_callback=lambda err, fail_count: (
                    _handle_deadly_error(err, fail_count),
                    io.seek(0, 0),
                    logger.warning(
                        "`upload_file`: `upload_slice`: error: %s, fail_count: %s",
                        err,
                        fail_count,
                        exc_info=err,
                    ),
                    _wait_start(),
                ),
            )(api.upload_slice)(io, callback=callback_for_slice)

            slice_md5s.append(slice_md5)
            slice_completed += size
            idx += 1

        # Combine slices
        _combine_slices(
            api,
            remotepath,
            slice_md5s,
            local_ctime,
            local_mtime,
            ondup,
        )

        logger.debug(
            "`upload_file`: upload_slice and combine_slices success, task_id: %s",
            task_id,
        )

        # `combine_slices` can not get right content md5.
        # We need to check whether server updates by hand.
        if check_md5:
            _check_md5(
                api,
                localpath,
                remotepath,
                slice256k_md5,
                content_md5,
                content_crc32,
                encrypt_io_len,
                encrypt_password=encrypt_password,
                encrypt_type=encrypt_type.value,
                user_id=user_id,
                user_name=user_name,
            )

        if task_id is not None and progress_task_exists(task_id):
            _progress.remove_task(task_id)
    except Exception as err:
        logger.warning("`upload_file`: error: %s", err)
        raise err
    finally:
        encrypt_io.close()
        if task_id is not None and progress_task_exists(task_id):
            _progress.reset(task_id)


def _check_md5(
    api: BaiduPCSApi,
    localpath: str,
    remotepath: str,
    slice_md5: str,
    content_md5: str,
    content_crc32: int,  # not needed
    content_length: int,
    encrypt_password: bytes = b"",
    encrypt_type: str = "",
    user_id: Optional[int] = None,
    user_name: Optional[str] = None,
):
    """Fix remote content md5 with rapid upload

    There is a delay for server to handle uploaded data after `combine_slices`,
    so we retry fix it.
    """

    i = 0
    while True:
        logger.debug(
            f"`_check_md5`: retry: {i}: "
            "slice_md5: %s, content_md5: %s, content_crc32: %s, io_len: %s, remotepath: %s",
            slice_md5,
            content_md5,
            content_crc32,
            content_length,
            remotepath,
        )
        i += 1

        try:
            api.rapid_upload_file(
                slice_md5,
                content_md5,
                content_crc32,  # not needed
                content_length,
                remotepath,
                ondup="overwrite",
            )
            logger.warning("`_check_md5`: successes")

            if _rapiduploadinfo_file:
                save_rapid_upload_info(
                    _rapiduploadinfo_file,
                    slice_md5,
                    content_md5,
                    content_crc32,
                    content_length,
                    localpath=localpath,
                    remotepath=remotepath,
                    encrypt_password=encrypt_password,
                    encrypt_type=encrypt_type,
                    user_id=user_id,
                    user_name=user_name,
                )
            return
        except Exception as err:
            logger.warning("`_check_md5`: fails: %s", err)
            time.sleep(2)
            continue



from io import BytesIO
from typing import IO, Callable, Dict, List, Optional, Tuple

from PIL import Image
from requests_toolbelt import MultipartEncoderMonitor
from rich import print
from rich.prompt import Prompt

from ..common import constant
from ..common.crypto import calu_md5
from .api.common.io import MAX_CHUNK_SIZE, RangeRequestIO
from .inner import (
    CloudTask,
    FromTo,
    PcsAuth,
    PcsFile,
    PcsMagnetFile,
    PcsQuota,
    PcsRapidUploadInfo,
    PcsSharedLink,
    PcsSharedPath,
    PcsUser,
    PcsUserProduct,
)
from .pcs import BaiduPCS, BaiduPCSError, M3u8Type


class BaiduPCSApi:
    """Baidu PCS Api

    This is the wrapper of `BaiduPCS`. It parses the content of response of raw
    BaiduPCS requests to some inner data structions.
    """

    def __init__(
        self,
        bduss: Optional[str] = None,
        stoken: Optional[str] = None,
        ptoken: Optional[str] = None,
        cookies: Dict[str, Optional[str]] = {},
        user_id: Optional[int] = None,
    ):
        self._baidupcs = BaiduPCS(
            bduss, stoken=stoken, ptoken=ptoken, cookies=cookies, user_id=user_id
        )

    @property
    def bduss(self) -> str:
        return self._baidupcs._bduss

    @property
    def bdstoken(self) -> str:
        return self._baidupcs.bdstoken

    @property
    def stoken(self) -> Optional[str]:
        return self._baidupcs._stoken

    @property
    def ptoken(self) -> Optional[str]:
        return self._baidupcs._ptoken

    @property
    def baiduid(self) -> Optional[str]:
        return self._baidupcs._baiduid

    @property
    def logid(self) -> Optional[str]:
        return self._baidupcs._logid

    @property
    def user_id(self) -> Optional[int]:
        return self._baidupcs._user_id

    @property
    def cookies(self) -> Dict[str, Optional[str]]:
        return self._baidupcs.cookies

    def quota(self) -> PcsQuota:
        """Quota Information"""

        info = self._baidupcs.quota()
        return PcsQuota(quota=info["quota"], used=info["used"])

    def meta(self, *remotepaths: str) -> List[PcsFile]:
        """Meta data of `remotepaths`"""

        info = self._baidupcs.meta(*remotepaths)
        return [PcsFile.from_(v) for v in info.get("list", [])]

    def exists(self, remotepath: str) -> bool:
        """Check whether `remotepath` exists"""

        return self._baidupcs.exists(remotepath)

    def is_file(self, remotepath: str) -> bool:
        """Check whether `remotepath` is a file"""

        return self._baidupcs.is_file(remotepath)

    def is_dir(self, remotepath: str) -> bool:
        """Check whether `remotepath` is a directory"""

        return self._baidupcs.is_dir(remotepath)

    def list(
        self,
        remotepath: str,
        desc: bool = False,
        name: bool = False,
        time: bool = False,
        size: bool = False,
        recursive: bool = False,
    ) -> List[PcsFile]:
        """List directory contents"""

        info = self._baidupcs.list(
            remotepath, desc=desc, name=name, time=time, size=size
        )
        pcs_files = [PcsFile.from_(v) for v in info.get("list", [])]
        if recursive:
            for pcs_file in pcs_files:
                if pcs_file.is_dir:
                    sub_pcs_files = self.list(
                        pcs_file.path, desc=desc, name=name, time=time, size=size
                    )
                    pcs_files.extend(sub_pcs_files)
        return pcs_files

    def upload_file(
        self,
        io: IO,
        remotepath: str,
        ondup="overwrite",
        callback: Callable[[MultipartEncoderMonitor], None] = None,
    ) -> PcsFile:
        """Upload an io to `remotepath`

        ondup (str): "overwrite" or "newcopy"
        callable: the callback for monitoring uploading progress

        Warning, the api CAN NOT set local_ctime and local_mtime
        """

        info = self._baidupcs.upload_file(
            io, remotepath, ondup=ondup, callback=callback
        )
        return PcsFile.from_(info)

    def rapid_upload_file(
        self,
        slice_md5: str,
        content_md5: str,
        content_crc32: int,  # not needed
        io_len: int,
        remotepath: str,
        local_ctime: Optional[int] = None,
        local_mtime: Optional[int] = None,
        ondup="overwrite",
    ) -> PcsFile:
        """Rapid Upload File

        slice_md5 (32 bytes): the md5 of pre 256KB of content.
        content_md5 (32 bytes): the md5 of total content.
        content_crc32 (int): the crc32 of total content (Not Needed),
            if content_crc32 is 0, the params of the api will be ignored.
        io_len (int): the length of total content.
        remotepath (str): the absolute remote path to save the content.
        local_ctime (optional, int): the timestramp of the local ctime
        local_mtime (optional, int): the timestramp of the local mtime
        ondup (str): "overwrite" or "newcopy"
        """

        info = self._baidupcs.rapid_upload_file(
            slice_md5,
            content_md5,
            content_crc32,
            io_len,
            remotepath,
            local_ctime=local_ctime,
            local_mtime=local_mtime,
            ondup=ondup,
        )
        return PcsFile.from_(info)

    def upload_slice(
        self, io: IO, callback: Callable[[MultipartEncoderMonitor], None] = None
    ) -> str:
        """Upload an io as a slice

        callable: the callback for monitoring uploading progress
        """

        info = self._baidupcs.upload_slice(io, callback=callback)
        return info["md5"]

    def combine_slices(
        self,
        slice_md5s: List[str],
        remotepath: str,
        local_ctime: Optional[int] = None,
        local_mtime: Optional[int] = None,
        ondup="overwrite",
    ) -> PcsFile:
        """Combine uploaded slices to `remotepath`

        local_ctime (optional, int): the timestramp of the local ctime
        local_mtime (optional, int): the timestramp of the local mtime
        ondup (str): "overwrite" or "newcopy"
        """

        info = self._baidupcs.combine_slices(
            slice_md5s,
            remotepath,
            local_ctime=local_ctime,
            local_mtime=local_mtime,
            ondup=ondup,
        )
        return PcsFile.from_(info)

    def search(
        self, keyword: str, remotepath: str, recursive: bool = False
    ) -> List[PcsFile]:
        """Search in `remotepath` with `keyword`"""

        info = self._baidupcs.search(keyword, remotepath, recursive=recursive)
        pcs_files = []
        for file_info in info["list"]:
            pcs_files.append(PcsFile.from_(file_info))
        return pcs_files

    def makedir(self, directory: str) -> PcsFile:
        info = self._baidupcs.makedir(directory)
        return PcsFile.from_(info)

    def move(self, *remotepaths: str) -> List[FromTo]:
        """Move `remotepaths[:-1]` to `remotepaths[-1]`"""

        info = self._baidupcs.move(*remotepaths)
        r = info["extra"].get("list")
        if not r:
            raise BaiduPCSError("File operator [move] fails")
        return [FromTo(from_=v["from"], to_=v["to"]) for v in r]

    def rename(self, source: str, dest: str) -> FromTo:
        info = self._baidupcs.rename(source, dest)
        r = info["extra"].get("list")
        if not r:
            raise BaiduPCSError("File operator [rename] fails")
        v = r[0]
        return FromTo(from_=v["from"], to_=v["to"])

    def copy(self, *remotepaths: str):
        """Copy `remotepaths[:-1]` to `remotepaths[-1]`"""

        info = self._baidupcs.copy(*remotepaths)
        r = info["extra"].get("list")
        if not r:
            raise BaiduPCSError("File operator [copy] fails")
        return [FromTo(from_=v["from"], to_=v["to"]) for v in r]

    def remove(self, *remotepaths: str):
        """Remove all `remotepaths`"""

        self._baidupcs.remove(*remotepaths)

    def magnet_info(self, magnet: str) -> List[PcsMagnetFile]:
        """Get the magnet information"""

        info = self._baidupcs.magnet_info(magnet)
        return [PcsMagnetFile.from_(v) for v in info["magnet_info"]]

    def torrent_info(self, remote_torrent: str):
        """Get the `remote_torrent` information"""

        self._baidupcs.torrent_info(remote_torrent)

    def add_task(self, task_url: str, remotedir: str) -> str:
        """Add a cloud task to save at `remotedir`

        task_url (str): http url
        """

        info = self._baidupcs.add_task(task_url, remotedir)
        return str(info["task_id"])

    def add_magnet_task(
        self, task_url: str, remotedir: str, selected_idx: List[int]
    ) -> str:
        """Add a magnet task to save at `remotedir`.

        task_url (str): magnet link
        selected_idx: the indexes needed to download
        """

        info = self._baidupcs.add_magnet_task(task_url, remotedir, selected_idx)
        return str(info["task_id"])

    def tasks(self, *task_ids: str) -> List[CloudTask]:
        """List cloud tasks with their `task_ids`"""

        info = self._baidupcs.tasks(*task_ids)
        tasks = []
        for task_id, v in info["task_info"].items():
            v["task_id"] = task_id
            tasks.append(CloudTask.from_(v))
        return tasks

    def list_tasks(self) -> List[CloudTask]:
        """List all cloud tasks"""

        info = self._baidupcs.list_tasks()
        return [CloudTask.from_(v) for v in info["task_info"]]

    def clear_tasks(self) -> int:
        """Clear all finished and failed cloud tasks"""

        info = self._baidupcs.clear_tasks()
        return info["total"]

    def cancel_task(self, task_id: str):
        """Cancel a cloud task with its `task_id`"""

        self._baidupcs.cancel_task(task_id)

    def share(self, *remotepaths: str, password: str, period: int = 0) -> PcsSharedLink:
        """Share `remotepaths` to public with a optional password

        To use api, `STOKEN` must be in `cookies`

        period (int): The days for expiring. `0` means no expiring
        """

        info = self._baidupcs.share(*remotepaths, password=password, period=period)
        link = PcsSharedLink.from_(info)._replace(
            paths=list(remotepaths), password=password
        )
        return link

    def list_shared(self, page: int = 1) -> List[PcsSharedLink]:
        """List shared link on a page

        To use api, `STOKEN` must be in `cookies`
        """

        info = self._baidupcs.list_shared(page)
        return [PcsSharedLink.from_(v) for v in info["list"]]

    def shared_password(self, share_id: int) -> Optional[str]:
        """Show shared link password

        To use api, `STOKEN` must be in `cookies`
        """

        info = self._baidupcs.shared_password(share_id)
        p = info.get("pwd", "0")  # If "pwd" is not in info, error is 分享已过期
        if p == "0":
            return None
        return p

    def cancel_shared(self, *share_ids: int):
        """Cancel shared links with their `share_ids`

        To use api, `STOKEN` must be in `cookies`
        """

        self._baidupcs.cancel_shared(*share_ids)

    def access_shared(
        self,
        shared_url: str,
        password: str,
        vcode_str: str = "",
        vcode: str = "",
        show_vcode: bool = True,
    ):
        """Verify the `shared_url` which needs the `password`

        This method MUST be called before calling `self.shared_paths`.

        show_vcode (bool): If set True, it will be open the vcode image (if needed)
            with a gui window. Else, you need to handle the vcode.
        """

        while True:
            try:
                self._baidupcs.access_shared(shared_url, password, vcode_str, vcode)
                return
            except BaiduPCSError as err:
                if err.error_code not in (-9, -62):
                    raise err
                if show_vcode:
                    if err.error_code == -62:  # -62: '可能需要输入验证码'
                        print("[yellow]Need vcode![/yellow]")
                    if err.error_code == -9:
                        print("[red]vcode is incorrect![/red]")
                    vcode_str, vcode_img_url = self.getcaptcha(shared_url)
                    img_cn = self.get_vcode_img(vcode_img_url, shared_url)
                    img_buf = BytesIO(img_cn)
                    img_buf.seek(0, 0)
                    img = Image.open(img_buf)
                    img.show()
                    vcode = Prompt.ask("input vcode")
                else:
                    raise err

    def getcaptcha(self, shared_url: str) -> Tuple[str, str]:
        """Get one vcode information

        Return `vcode_str`, `vcode_img_url`"""

        info = self._baidupcs.getcaptcha(shared_url)
        return info["vcode_str"], info["vcode_img"]

    def get_vcode_img(self, vcode_img_url: str, shared_url: str) -> bytes:
        """Get vcode image content"""

        return self._baidupcs.get_vcode_img(vcode_img_url, shared_url)

    def shared_paths(self, shared_url: str) -> List[PcsSharedPath]:
        """Shared paths of the `shared_url`"""

        info = self._baidupcs.shared_paths(shared_url)
        uk = info.get("share_uk") or info.get("uk")
        uk = int(uk)

        assert uk, "`BaiduPCSApi.shared_paths`: Don't get `uk`"

        share_id = info["shareid"]
        bdstoken = info["bdstoken"]

        if not info.get("file_list"):
            return []

        if isinstance(info["file_list"], list):
            file_list = info["file_list"]
        elif isinstance(info["file_list"].get("list"), list):
            file_list = info["file_list"]["list"]
        else:
            raise ValueError("`shared_paths`: Parsing shared info fails")

        return [
            PcsSharedPath.from_(v)._replace(uk=uk, share_id=share_id, bdstoken=bdstoken)
            for v in file_list
        ]

    def list_shared_paths(
        self,
        sharedpath: str,
        uk: int,
        share_id: int,
        bdstoken: str,
        page: int = 1,
        size: int = 100,
    ) -> List[PcsSharedPath]:
        """Sub shared paths of the shared directory `sharedpath`"""

        info = self._baidupcs.list_shared_paths(
            sharedpath, uk, share_id, page=page, size=size
        )
        return [
            PcsSharedPath.from_(v)._replace(uk=uk, share_id=share_id, bdstoken=bdstoken)
            for v in info["list"]
        ]

    def transfer_shared_paths(
        self,
        remotedir: str,
        fs_ids: List[int],
        uk: int,
        share_id: int,
        bdstoken: str,
        shared_url: str,
    ):
        """Save these `fs_ids` of shared paths to `remotedir`"""

        self._baidupcs.transfer_shared_paths(
            remotedir, fs_ids, uk, share_id, bdstoken, shared_url
        )

    def user_info(self) -> PcsUser:
        """User's information"""

        info = self._baidupcs.user_info()
        user_id = int(info["user"]["id"])
        user_name = info["user"]["name"]

        info = self._baidupcs.tieba_user_info(user_id)
        age = float(info["user"]["tb_age"])
        sex = info["user"]["sex"]
        if sex == 1:
            sex = "♂"
        elif sex == 2:
            sex = "♀"
        else:
            sex = "unknown"

        auth = PcsAuth(
            bduss=self._baidupcs._bduss,
            cookies=self.cookies,
            stoken=self._baidupcs._stoken,
            ptoken=self._baidupcs._ptoken,
        )

        quota = self.quota()

        products, level = self.user_products()

        return PcsUser(
            user_id=user_id,
            user_name=user_name,
            auth=auth,
            age=age,
            sex=sex,
            quota=quota,
            products=products,
            level=level,
        )

    def user_products(self) -> Tuple[List[PcsUserProduct], int]:
        """User's product information"""

        info = self._baidupcs.user_products()
        pds = []
        for p in info["product_infos"]:
            # `product_name` of some entries are None (issue #30)
            if not p.get("product_name"):
                continue

            pds.append(
                PcsUserProduct(
                    name=p["product_name"],
                    start_time=p["start_time"],
                    end_time=p["end_time"],
                )
            )

        level = info["level_info"]["current_level"]
        return pds, level

    def download_link(self, remotepath: str, pcs: bool = False) -> Optional[str]:
        """Download link of the `remotepath`

        pcs (bool, default: False): If pcs is True, return the downloading pcs link
            which has a limited threshold of downstream even if the user is a svip.
            If pcs is False, return the downloading link requested by the android api
            and which has not limited threshold for a svip user.
        """

        return self._baidupcs.download_link(remotepath, pcs=pcs)

    def file_stream(
        self,
        remotepath: str,
        max_chunk_size: int = MAX_CHUNK_SIZE,
        callback: Callable[..., None] = None,
        encrypt_password: bytes = b"",
        pcs: bool = False,
    ) -> Optional[RangeRequestIO]:
        """File stream as a normal io"""

        return self._baidupcs.file_stream(
            remotepath,
            max_chunk_size=max_chunk_size,
            callback=callback,
            encrypt_password=encrypt_password,
            pcs=pcs,
        )

    def m3u8_stream(self, remotepath: str, type: M3u8Type = "M3U8_AUTO_720") -> str:
        """Media file's m3u8 content"""

        info = self._baidupcs.m3u8_stream(remotepath, type)
        if info.get("m3u8_content"):
            return info["m3u8_content"]
        else:
            # Here should be a error
            return ""

    def rapid_upload_info(
        self, remotepath: str, check: bool = True
    ) -> Optional[PcsRapidUploadInfo]:
        """Rapid upload information

        check (bool): If check is True, we need to use the `self.rapid_upload_file` to
            check whether the rapid upload information is valid. Therefore some meta information
            of the `remotepath` will be changed. These are server_ctime and server_mtime.
        """

        pcs_file = self.meta(remotepath)[0]
        content_length = pcs_file.size or 0

        if content_length < 256 * constant.OneK:
            return None

        fs = self.file_stream(remotepath, pcs=False)
        if not fs:
            return None

        data = fs.read(256 * constant.OneK)
        assert data and len(data) == 256 * constant.OneK

        slice_md5 = calu_md5(data)

        assert (
            content_length and content_length == fs._auto_decrypt_request.content_length
        )

        content_md5 = fs._auto_decrypt_request.content_md5
        content_crc32 = fs._auto_decrypt_request.content_crc32 or 0

        if not content_md5:
            return None

        block_list = pcs_file.block_list
        if block_list and len(block_list) == 1 and block_list[0] == pcs_file.md5:
            return PcsRapidUploadInfo(
                slice_md5=slice_md5,
                content_md5=content_md5,
                content_crc32=content_crc32,
                content_length=content_length,
                remotepath=pcs_file.path,
            )

        if check:
            try:
                # Try rapid_upload_file
                self.rapid_upload_file(
                    slice_md5,
                    content_md5,
                    content_crc32,
                    content_length,
                    pcs_file.path,
                    local_ctime=pcs_file.local_ctime,
                    local_mtime=pcs_file.local_mtime,
                    ondup="overwrite",
                )
            except BaiduPCSError as err:
                # 31079: "未找到文件MD5"
                if err.error_code != 31079:
                    raise err
                return None

        return PcsRapidUploadInfo(
            slice_md5=slice_md5,
            content_md5=content_md5,
            content_crc32=content_crc32,
            content_length=content_length,
            remotepath=pcs_file.path,
        )



from typing import Optional

import chardet  # type: ignore

from ..common.io import MAX_CHUNK_SIZE
from ..pcs import BaiduPCSApi
from .display import display_blocked_remotepath


def cat(
    api: BaiduPCSApi,
    remotepath: str,
    max_chunk_size: int = MAX_CHUNK_SIZE,
    encoding: Optional[str] = None,
    encrypt_password: bytes = b"",
):
    fs = api.file_stream(remotepath, encrypt_password=encrypt_password)
    if not fs:
        display_blocked_remotepath(remotepath)
        return

    cn = fs.read()
    if cn:
        if encoding:
            print(cn.decode(encoding))
        else:
            r = chardet.detect(cn)
            if r["confidence"] > 0.5:
                print(cn.decode(r["encoding"]))  # type: ignore
            else:
                print(cn)



import os
import re
from collections import deque
from pathlib import Path, PurePosixPath
from typing import Dict, List, Optional, Set

from rich import print

from ..pcs import BaiduPCSApi, BaiduPCSError, PcsSharedPath
from .display import display_shared_links, display_shared_paths

SHARED_URL_PREFIX = "https://pan.baidu.com/s/"


def _unify_shared_url(url: str) -> str:
    """Unify input shared url"""

    # For Standard url
    temp = r"pan\.baidu\.com/s/(.+?)(\?|$)"
    m = re.search(temp, url)
    if m:
        return SHARED_URL_PREFIX + m.group(1)

    # For surl url
    temp = r"baidu\.com.+?\?surl=(.+?)(\?|$)"
    m = re.search(temp, url)
    if m:
        return SHARED_URL_PREFIX + "1" + m.group(1)

    raise ValueError(f"The shared url is not a valid url. {url}")


def share_files(api: BaiduPCSApi, *remotepaths: str, password: str, period: int = 0):
    shared_link = api.share(*remotepaths, password=password, period=period)
    display_shared_links(shared_link)


def list_shared(api: BaiduPCSApi, show_all=True):
    page = 1
    while True:
        shared_links = api.list_shared(page=page)
        if not shared_links:
            break

        page += 1

        for sl in shared_links:
            if sl.has_password():
                assert sl.share_id
                pwd = api.shared_password(sl.share_id)
                sl = sl._replace(password=pwd)

            if show_all or sl.available():
                display_shared_links(sl)


def cancel_shared(api: BaiduPCSApi, *share_ids: int):
    api.cancel_shared(*share_ids)


def save_shared(
    api: BaiduPCSApi,
    shared_url: str,
    remotedir: str,
    password: Optional[str] = None,
    show_vcode: bool = True,
):
    assert remotedir.startswith("/"), "`remotedir` must be an absolute path"

    shared_url = _unify_shared_url(shared_url)

    # Vertify with password
    if password:
        api.access_shared(shared_url, password, show_vcode=show_vcode)

    shared_paths = deque(api.shared_paths(shared_url))

    # Record the remotedir of each shared_path
    _remotedirs: Dict[PcsSharedPath, str] = {}
    for sp in shared_paths:
        _remotedirs[sp] = remotedir

    _dir_exists: Set[str] = set()

    while shared_paths:
        shared_path = shared_paths.popleft()
        rd = _remotedirs[shared_path]

        # Make sure remote dir exists
        if rd not in _dir_exists:
            if not api.exists(rd):
                api.makedir(rd)
            _dir_exists.add(rd)

        # Ignore existed file
        if shared_path.is_file and remotepath_exists(
            api, PurePosixPath(shared_path.path).name, rd
        ):
            print(f"[yellow]WARNING[/]: {shared_path.path} has be in {rd}")
            continue

        uk, share_id, bdstoken = (
            shared_path.uk,
            shared_path.share_id,
            shared_path.bdstoken,
        )
        assert uk
        assert share_id
        assert bdstoken

        try:
            api.transfer_shared_paths(
                rd, [shared_path.fs_id], uk, share_id, bdstoken, shared_url
            )
            print(f"save: {shared_path.path} to {rd}")
            continue
        except BaiduPCSError as err:
            if err.error_code == 12:  # 12: "文件已经存在"
                print(
                    f"[yellow]WARNING[/]: error_code: {err.error_code}, {shared_path.path} has be in {rd}"
                )
            elif err.error_code == -32:  # -32: "剩余空间不足，无法转存",
                raise err
            elif err.error_code in (
                -33,  # -33: "一次支持操作999个，减点试试吧"
                4,  # 4: "share transfer pcs error"
                130,  # "转存文件数超限"
                120,  # "转存文件数超限"
            ):
                print(
                    f"[yellow]WARNING[/]: error_code: {err.error_code}, {shared_path.path} "
                    "has more items and need to transfer one by one"
                )
            else:
                raise err

        if shared_path.is_dir:
            # Take all sub paths
            sub_paths = list_all_sub_paths(
                api, shared_path.path, uk, share_id, bdstoken
            )

            rd = (Path(rd) / os.path.basename(shared_path.path)).as_posix()
            for sp in sub_paths:
                _remotedirs[sp] = rd
            shared_paths.extendleft(sub_paths[::-1])


def list_all_sub_paths(
    api: BaiduPCSApi,
    sharedpath: str,
    uk: int,
    share_id: int,
    bdstoken: str,
) -> List[PcsSharedPath]:
    sub_paths = []
    page = 1
    size = 100
    while True:
        sps = api.list_shared_paths(
            sharedpath, uk, share_id, bdstoken, page=page, size=size
        )
        sub_paths += sps
        if len(sps) < 100:
            break
        page += 1
    return sub_paths


def list_shared_paths(
    api: BaiduPCSApi,
    shared_url: str,
    password: Optional[str] = None,
    show_vcode: bool = True,
):
    shared_url = _unify_shared_url(shared_url)

    # Vertify with password
    if password:
        api.access_shared(shared_url, password, show_vcode=show_vcode)

    all_shared_paths: List[PcsSharedPath] = []

    shared_paths = deque(api.shared_paths(shared_url))
    all_shared_paths += shared_paths

    while shared_paths:
        shared_path = shared_paths.popleft()

        uk, share_id, bdstoken = (
            shared_path.uk,
            shared_path.share_id,
            shared_path.bdstoken,
        )
        assert uk
        assert share_id
        assert bdstoken

        if shared_path.is_dir:
            # Take all sub paths
            sub_paths = list_all_sub_paths(
                api, shared_path.path, uk, share_id, bdstoken
            )
            all_shared_paths += sub_paths
            shared_paths.extendleft(sub_paths[::-1])

    display_shared_paths(*all_shared_paths)


def remotepath_exists(
    api: BaiduPCSApi, name: str, rd: str, _cache: Dict[str, Set[str]] = {}
) -> bool:
    names = _cache.get(rd)
    if not names:
        names = set([PurePosixPath(sp.path).name for sp in api.list(rd)])
        _cache[rd] = names
    return name in names



from typing import List

from rich.console import Console

from ..pcs import BaiduPCSApi
from ..utils import human_size
from .log import get_logger
from .sifter import Sifter, sift

logger = get_logger(__name__)

DEFAULT_MAX_WORKERS = 10


def du(
    api: BaiduPCSApi,
    remotepath: str,
    recursive: bool = False,
    sifters: List[Sifter] = [],
    highlight: bool = False,
) -> int:
    is_dir = api.is_dir(remotepath)
    if is_dir:
        pcs_files = api.list(remotepath)
    else:
        pcs_files = api.meta(remotepath)

    pcs_files = sift(pcs_files, sifters, recursive=recursive)
    if not pcs_files:
        return 0

    usage = 0
    for pcs_file in pcs_files:
        if pcs_file.is_dir:
            if recursive:
                usage += du(
                    api,
                    pcs_file.path,
                    recursive=recursive,
                    sifters=sifters,
                    highlight=highlight,
                )
        else:
            usage += pcs_file.size or 0

    return usage


def disk_usage(
    api: BaiduPCSApi,
    *remotepaths: str,
    recursive: bool = False,
    sifters: List[Sifter] = [],
):
    console = Console()
    with console.status("") as status:
        for rp in remotepaths:
            status.update(status=f"[b green]Request remote[/b green]: {rp}")
            usage = du(api, rp, recursive=recursive, sifters=sifters)
            console.print(f"{human_size(usage)}\t{usage}\t{rp}")



from typing import Optional, List

from ..pcs import BaiduPCSApi
from .display import display_files
from .sifter import Sifter, IncludeSifter


def search(
    api: BaiduPCSApi,
    keyword: str,
    remotedir: str = "/",
    recursive: bool = False,
    sifters: Optional[List[Sifter]] = None,
    highlight: bool = False,
    show_size: bool = False,
    show_date: bool = False,
    show_md5: bool = False,
    csv: bool = False,
):
    pcs_files = api.search(keyword, remotedir, recursive=recursive)

    sifters = [*(sifters or []), IncludeSifter(keyword)]
    display_files(
        pcs_files,
        remotedir,
        sifters=sifters,
        highlight=highlight,
        show_size=show_size,
        show_date=show_date,
        show_md5=show_md5,
        show_absolute_path=True,
        csv=csv,
    )



from typing import Optional, List, Tuple, Union
from pathlib import Path

from ..pcs import BaiduPCSApi, PcsFile, FromTo
from ..common.path import walk, join_path
from ..common.crypto import calu_file_md5
from ..common.constant import CPU_NUM
from ..common.io import EncryptType
from .upload import upload as _upload, DEFAULT_SLICE_SIZE
from .log import get_logger

from rich import print

logger = get_logger(__name__)


def recursive_list(api: BaiduPCSApi, remotedir: Union[str, PcsFile]) -> List[PcsFile]:
    if isinstance(remotedir, PcsFile):
        remotedir = remotedir.path

    pcs_files = []
    for pcs_file in api.list(remotedir):
        if pcs_file.is_file:
            pcs_files.append(pcs_file)
        else:
            pcs_files.extend(recursive_list(api, pcs_file.path))
    return pcs_files


def check_file_md5(localpath: str, pcs_file: PcsFile) -> bool:
    local_file_md5 = calu_file_md5(localpath)
    if local_file_md5 == pcs_file.md5:
        return True
    else:
        return False


def sync(
    api: BaiduPCSApi,
    localdir: str,
    remotedir: str,
    encrypt_password: bytes = b"",
    encrypt_type: EncryptType = EncryptType.No,
    max_workers: int = CPU_NUM,
    slice_size: int = DEFAULT_SLICE_SIZE,
    show_progress: bool = True,
    rapiduploadinfo_file: Optional[str] = None,
    user_id: Optional[int] = None,
    user_name: Optional[str] = None,
    check_md5: bool = False,
):
    localdir = Path(localdir).as_posix()
    remotedir = Path(remotedir).as_posix()

    is_file = api.is_file(remotedir)
    assert not is_file, "remotedir must be a directory"

    if not api.exists(remotedir):
        all_pcs_files = {}
    else:
        all_pcs_files = {
            pcs_file.path[len(remotedir) + 1 :]: pcs_file
            for pcs_file in recursive_list(api, remotedir)
        }

    fts: List[FromTo] = []
    check_list: List[Tuple[str, PcsFile]] = []
    all_localpaths = set()
    for localpath in walk(Path(localdir)):
        path = localpath[len(localdir) + 1 :]
        all_localpaths.add(path)

        if path not in all_pcs_files:
            fts.append(FromTo(localpath, join_path(Path(remotedir), Path(path))))
        else:
            check_list.append((localpath, all_pcs_files[path]))

    for lp, pf in check_list:
        lstat = Path(lp).stat()
        if int(lstat.st_mtime) != pf.local_mtime or lstat.st_size != pf.size:
            fts.append(FromTo(lp, pf.path))

    to_deletes = []
    for rp in all_pcs_files.keys():
        if rp not in all_localpaths:
            to_deletes.append(all_pcs_files[rp].path)

    logger.debug(
        "`sync`: all localpaths: %s, "
        "localpaths needed to upload: %s, "
        "remotepaths needed to delete: %s",
        len(all_localpaths),
        len(fts),
        len(to_deletes),
    )

    # The md5 of remote file is incorrect at most time, so we don't compare md5
    #
    # # Compare localpath content md5 with remotepath content md5
    # semaphore = Semaphore(max_workers)
    # with ThreadPoolExecutor(max_workers=CPU_NUM) as executor:
    #     tasks = {}
    #     for lp, pf in check_list:
    #         semaphore.acquire()
    #         fut = executor.submit(sure_release, semaphore, check_file_md5, lp, pf)
    #         tasks[fut] = (lp, pf)
    #
    #     for fut in as_completed(tasks):
    #         is_equal = fut.result()
    #         lp, pf = tasks[fut]
    #         if not is_equal:
    #             fts.append(FromTo(lp, pf.path))

    _upload(
        api,
        fts,
        encrypt_password=encrypt_password,
        encrypt_type=encrypt_type,
        max_workers=max_workers,
        slice_size=slice_size,
        ignore_existing=False,
        show_progress=show_progress,
        rapiduploadinfo_file=rapiduploadinfo_file,
        user_id=user_id,
        user_name=user_name,
        check_md5=check_md5,
    )

    if to_deletes:
        api.remove(*to_deletes)
        print(f"Delete: [i]{len(to_deletes)}[/i] remote paths")
